/*
 * Drood Copyright (c) 2007-2012, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Drood is a minimal, multiplexing, kqueue-based, partially-compliant
 * HTTP/1.1 server.  The only parts of HTTP/1.1 it implements are
 * persistent connections, the "chunked" transfer-encoding, the "gzip"
 * content-encoding, single byte-ranges, and name-based virtual hosting
 * dictated by the value of the "Host" request header.
 */

#include <stdlib.h>
#include <limits.h>
#include <inttypes.h>
#include <stdio.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <dirent.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <fcntl.h>
#include <signal.h>
#include <syslog.h>
#include <stdarg.h>

#include <string.h>
#include <ctype.h>

#define VERSION "3.3"
#define PIDFILE "/var/run/drood.pid"
#define TYPEFILE DATADIR "/types.tab"

/*
 * Size of the output queue used by kevent().
 */

#define QLEN 5000

/*
 * This is size of the IO buffers used in connection control blocks.
 * Modifying this value may require a corresponding modification to
 * the precision of the chunk header conversion performed in
 * respond_chunked().  32K is the default size of a socket send buffer
 * on i386 FreeBSD 6.x.  We try to move data in chunks of this size in an
 * attempt to "fill the pipe" of each connection.
 */

#define IOBUFSIZE 32768

/*
 * This is the size of the input buffers used to read request headers.  It
 * should not be altered.  This is the maximum amount we ever want to
 * over-read into the next request header.  If the next request is a POST,
 * we don't want to read into the next request's body.

 * The request line and a Content-Length header will be the minimum header
 * we accept for POSTS, therefore we are able to over-read 35 characters
 * into the next request header safely.  This assumes the client only
 * terminated lines with newlines and not carriage-return newline pairs,
 * just in case.  An over-read of 35 characters will occur when the first
 * buffered character is the final newline terminating the current request
 * header.  The rest of the buffer will be over-read characters from the
 * next request header.  Include the string terminator, and that's why
 * INBUFSIZE is 37 and not 35.

 * We use the "first" flag in ccbs to cause receive_request() to read the
 * request line character-by-character.  If the request method is not POST,
 * then we reset the flag and subsequent reads will be buffered.  If the
 * request is a POST, receive_request() continues to read characters
 * individually, so as not to read into the current request's body.  If
 * the -d option is not present on the command-line, we are not handling
 * (S)CGI requests, so in that case, flag is always 0.  This gives us a
 * little performance boost.

 * Buffering 36 characters reduces the number of read() system calls we
 * make for every request header.  Previously, we read each character
 * separately.
 */

#define INBUFSIZE 37

/*
 * A connection is in one of the following states at any one time.  There
 * are handlers for each state which are invoked when kevent() returns an
 * event on a connection.  A connection may have the handler for its
 * current state called more than once as part of the multiplexing process.
 * Eventually a handler will advance a connection to a succeeding state,
 * unless a system error occurs, or the connection times-out, when it will
 * be dropped.  Any connection in a state described as a terminal state,
 * will transition back to the RECEIVE state for a persistent connection,
 * otherwise the connection will be closed.

 * RECEIVE:  Receive request header.  Transitions to HEADER or ERROR.

 * HEADER:   Send response header.  Is terminal state for HEAD requests and
 *           (S)CGI redirects.  When a response body follows the header,
 *           transitions to RESPOND, NOCHUNK, CHUNK, or CONTINUE.

 * RESPOND:  Send response body for static resource,
 *           Invoked at least once for each IOBUFSIZE portion of requested
 *           resource.  Terminal state.

 * NOCHUNK:  Send a response body from a (S)CGI program which either
 *           provided a Content-Length header, or was invoked by an
 *           HTTP/1.0 request.  Invoked at least once for every IOBUFSIZE
 *           protion of the (S)CGI program's output.  Terminal state.

 * CHUNK:    Chunk a response body on-the-fly for an HTTP/1.1 (S)CGI
 *           request.  Invoked at least once for each IOBUFSIZE - 8
 *           portion of the (S)CGI program's output.  Terminal state.

 * CONTINUE: Continue processing a POST request which contained an
 *           "Expect: continue" header, after "100 Continue" has been sent.
 *           Transitions to FILTER or ERROR.

 * ERROR:    Send an error response header and body.  Terminal state.

 * FILTER:   Read and supplement a (S)CGI response header, as
 *           necessary.  Transitions to ERROR or HEADER.

 * SCGI_CONNECT:  Initiate non-blocking connect to SCGI server.
 *                Transitions to SGI_PREPARE or ERROR.

 * SCGI_PREPARE:  Verify non-blocking connect to SCGI server succeeded, and
 *                prepare the SCGI header for transmission.
 *                Transitions to SCGI_HEADER or ERROR.

 * SCGI_HEADER:   Send environment as netstring header to SCGI server.
 *                Transitions to SGI_BODY or FILTER.

 * SCGI_BODY:     Send client request body to SCGI server.  Transitions to
 *                FILTER.

 */

/*
 * Length of stat cache array.
 */

#define CACHE_SIZE 10

/*
 * connection control block.
 */

struct ccb
{
   int type, verno, error, close, expect, count, source,
      sock, inactive, timer, gzip, first;

   int request_found, close_found, expect_found, gzip_found,
      host_found, type_found, agent_found, since_found, length_found,
      unmodified_found, referrer_found, date_found, location_found,
      range_found;

   void ( *state )( struct ccb * ), ( *next )( struct ccb * );

   off_t size, written, scgi_len;

   char *request, *method, *agent, *cookie, *host, *length, *version, *addr,
      *since, *unmodified, *referer, *content, *scgi_host, *scgi_port,
      *range;

   char buffer[ IOBUFSIZE ], inbuf[ INBUFSIZE ], *inptr;
   struct stack *working;
};

struct sc
{
   char *file;
   time_t t;
   struct stat st;
} stat_cache[ CACHE_SIZE + 1 ];

void reset_signals();

void remove_conn( struct ccb * );

void receive_request( struct ccb * );
void analyze_request( struct ccb * );
void respond_static( struct ccb * );

void filter_cgi_header( struct ccb * );
void respond_chunked( struct ccb * );
void respond_unchunked( struct ccb * );

void respond_error( struct ccb * );
void respond_header( struct ccb * );

void prepare_dynamic( struct ccb * );
void connect_scgi( struct ccb * );
void prepare_scgi( struct ccb * );
void send_scgi_header( struct ccb * );
void send_scgi_body( struct ccb * );

void ( *RECEIVE  )( struct ccb * ) = receive_request;
void ( *RESPOND  )( struct ccb * ) = respond_static;

void ( *ERROR    )( struct ccb * ) = respond_error;
void ( *HEADER   )( struct ccb * ) = respond_header;

void ( *FILTER   )( struct ccb * ) = filter_cgi_header;
void ( *CHUNK    )( struct ccb * ) = respond_chunked;
void ( *NOCHUNK  )( struct ccb * ) = respond_unchunked;
void ( *CONTINUE )( struct ccb * ) = prepare_dynamic;

void ( *SCGI_CONNECT  )( struct ccb * ) = connect_scgi;
void ( *SCGI_PREPARE  )( struct ccb * ) = prepare_scgi;
void ( *SCGI_HEADER   )( struct ccb * ) = send_scgi_header;
void ( *SCGI_BODY     )( struct ccb * ) = send_scgi_body;

/*
 * These must start at 1.  0 means "no error".
 */

#define ERR_BADVERSION  1
#define ERR_BADREQUEST  2
#define ERR_FORBIDDEN   3
#define ERR_LENGTHREQ   4
#define ERR_BADMETHOD   5
#define ERR_NOTFOUND    6
#define ERR_INTERNAL    7
#define ERR_NOCONTENT   8
#define ERR_NOTMODIFIED 9
#define ERR_MODIFIED    10
#define ERR_RANGE       11

/*
 * KQueue Input and output queues.  We can add a maximun of four events to
 * the input queue for every item in the output queue.
 */

#define INQ QLEN * 4
struct kevent inqueue[ INQ ];
struct kevent outqueue[ QLEN ];

int in = 0, out = 0, closed = 0, read_timeout = 20, write_timeout = 60,
   instances = 0, scgi = 0, logfd = -1, active = 0, alloc_err = 0;

volatile int restart = 0, terminate = 0, sigchld = 0;

extern char *optarg;
char *root_resource = "index.html",
     *server_header = "Server: drood/" VERSION "\r\n",
     **g_argv = NULL;

int fd = -1, increment = 1024, root_resource_len = 0,
   logging = 0, testing = 0, backlog = 1000, dynalen = 0;

struct passwd *passwd;
struct group *group;

char *interface = "", *port = "80",
   *dynamic = NULL, *clean_dynamic = NULL,
   *grp = "nobody", *user = "nobody", *root = NULL,
   *default_type = "text/plain; charset=utf-8",
   *default_content_type = "Content-Type: text/plain; charset=utf-8\r\n",
   *keyfile = NULL, *certfile = NULL,
   *logfile = NULL;

struct stack *scgi_servers = NULL;
pid_t *children = NULL;

char *errstrings[ 12 ] = { "",
     " 505 HTTP Version Not Supported\r\n",
     " 400 Bad Request\r\n",
     " 403 Forbidden\r\n",
     " 411 Length Required\r\n",
     " 405 Method Not Allowed\r\nAllow: GET, HEAD, POST\r\n",
     " 404 Not Found\r\n",
     " 500 Internal Server Error\r\n",
     " 204 No Content\r\n",
     " 304 Not Modified\r\n",
     " 402 Precondition Failed\r\n",
     " 416 Request range not satisfiable\r\n"
};

#define STACK_PUSH( _s_, _o_ ) { if (( _s_ )->free )\
   {\
      if (( _s_ )->used ) ++( _s_ )->top;\
      --( _s_ )->free;\
      ++( _s_ )->used;\
      *( _s_ )->top = ( _o_ );\
   }\
   else stack_push( _s_, _o_ ); }

#define STRING_APPEND( _s_, _c_ ) { if (( _s_ )->free )\
   {\
      --( _s_ )->free;\
      ++( _s_ )->used;\
      *( _s_ )->top++ = _c_;\
      *( _s_ )->top ='\0';\
   }\
   else string_append( _s_, _c_ ); }

struct stack *type_table = NULL;

/*
 * String which grows dynamically.
 */

struct string
{
   int free, used;
   char *top, *str;
};

struct stack
{
   int free, used;
   char **top, **values;
};

void ev_set( int desc, short filter, u_short flags, struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = desc;
   kev->filter = filter;
   kev->fflags = 0;
   kev->flags = flags;
   kev->udata = item;
}

void flip_events( struct ccb *item )
{
   if ( item->source >= 0 )
      ev_set( item->source, EVFILT_READ, EV_DISABLE, item );

   ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );
   ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
}

void delete_timer( struct ccb *item )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = 0;
   kev->flags = EV_DELETE;
   kev->udata = item;
}

void set_timer( struct ccb *item, int read )
{
   struct kevent *kev;

   if ( in >= INQ )
      return;

   kev = &inqueue[ in++ ];

   kev->ident = item->sock;
   kev->filter = EVFILT_TIMER;
   kev->fflags = 0;
   kev->data = ( read ? read_timeout * 1000 : write_timeout * 60000 );
   kev->flags = EV_ADD;
   kev->udata = item;
}

void non_blocking( int desc )
{
   int flags, unblocked;

   if (( flags = fcntl( desc, F_GETFL, 0 )) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      if ( errno != EBADF )
         exit( 1 );
   }

   unblocked = flags & O_NONBLOCK;

   if ( ! unblocked && fcntl( desc, F_SETFL, flags | O_NONBLOCK ) < 0 )
   {
      syslog( LOG_ERR, "fcntl(): %m" );
      if ( errno != EBADF )
         exit( 1 );
   }
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "drood: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

struct stack *make_stack()
{
   struct stack *a;

   alloc_err = 1;
   
   if (( a = ( struct stack *)memory( sizeof( struct stack ))) == NULL )
      return NULL;

   if (( a->values = memory( sizeof( char * ) * increment )) == NULL )
      return NULL;

   alloc_err = 0;
   a->free = increment;
   a->used = 0;
   a->top = a->values;

   return a;
}

void stack_free( struct stack *a )
{
   free( a->values );
   free( a );
}

void stack_push( struct stack *a, char *o )
{
   if ( a->free == 0 )
   {
      alloc_err = 0;
      a->values = realloc( a->values, sizeof( char * ) * ( a->used + increment ) );

      if ( a->values == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "drood: realloc(): %s\n", strerror( errno ));

         alloc_err = 1;
         return;
      }

      a->free = increment;
      a->top = &a->values[ a->used - 1 ];
   }

   if ( a->used )
      ++a->top;

   *a->top = o;
   --a->free;
   ++a->used;
}

char *stack_pop( struct stack *a )
{
   char *ptr;

   if ( a->used )
   {
      ptr = *a->top;
      --a->used;
      ++a->free;

      if ( a->used )
         --a->top;

      return ptr;
   }

   return NULL;
}

char *str_dup( char *str, int len )
{
   char *ptr;

   if (( ptr = ( char *)memory( len + 1 )) != NULL )
   {
      bcopy( str, ptr, len );
      ptr[ len ] = '\0';
   }

   return ptr;
}

struct string *make_string()
{
   struct string *s;

   alloc_err = 1;

   s = ( struct string *)memory( sizeof( struct string ));
   if ( s == NULL )
      return s;

   alloc_err = 0;

   s->str = ( char *)memory( increment + 1 );
   *s->str = '\0';
   s->free = increment;
   s->used = 0;
   s->top = s->str;

   return s;
}

void string_free( struct string *s )
{
   free( s->str );
   free( s );
}

void signal_handler( int signo )
{
   if ( restart )
      return;

   /*
    * Avoid sending multiple signals to children if we get another
    * signal before the server has exited.
    */

   if ( ! terminate && children != NULL )
      while( instances > 0 )
         kill( children[ --instances ], SIGTERM );

   terminate = 1;
}

void sighup_handler( int signo )
{
   if ( terminate )
      return;

   if ( ! restart && children != NULL )
      while( instances > 0 )
         kill( children[ --instances ], SIGTERM );

   restart = 1;
}

void restart_drood()
{
   closelog();
   execv( g_argv[ 0 ], g_argv );
   syslog( LOG_ERR, "execvp(): %m" );
   exit( 1 );
}

void close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( %d, F_SETFD, FD_CLOEXEC ): %m", fd );
}

void string_append( struct string *s, char c )
{
   if ( s->free == 0 )
   {
      alloc_err = 0;
      s->str = ( char *)realloc( s->str, s->used + 1 + increment );

      if ( s->str == NULL )
      {
         if ( logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "drood: realloc(): %s\n", strerror( errno ));

         alloc_err = 1;
         return;
      }

      /* Leave room for end-of-string sentinel */

      s->free = increment;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';

   return;
}

void set_types()
{
   char *ptr, *dup, buffer[ 512 ];
   int fd, count;
   struct string *item;

   if (( type_table = make_stack()) == NULL )
      exit( 1 );

   if (( item = make_string() ) == NULL )
      exit( 1 );

   if (( fd = open( TYPEFILE, O_RDONLY )) < 0 )
   {
      fprintf( stderr, "drood: could not open %s\n", TYPEFILE );
      exit( 1 );
   }

   *buffer = '\0';
   ptr = buffer;

   for( ; ; )
   {
      if ( ! *ptr )
      {
         if (( count = read( fd, buffer, sizeof( buffer ) - 1 )) < 0 )
         {
            fprintf( stderr, "drood: could not read from %s: %s\n",
                     TYPEFILE, strerror( errno ));
            exit( 1 );
         }
         else if ( ! count )
         {
            if ( item->used )
               goto BREAK;

            break;
         }

         buffer[ count ] = '\0';
         ptr = buffer;
      }

      for( ; *ptr; ++ptr )
         switch( *ptr )
         {
            case '\t':
               while( *ptr == '\t' )
                  ++ptr;
               goto BREAK;

            case '\n':
               while( *ptr == '\n' )
                  ++ptr;
               goto BREAK;

            case '\r':
               continue;

            default:
               STRING_APPEND( item, *ptr )
               if ( alloc_err )
                  exit( 1 );
         }

      if ( ! *ptr )
         continue;

      if ( ! item->used )
      {
         fprintf( stderr, "drood: empty field in %s\n", TYPEFILE );
         exit( 1 );
      }

   BREAK:
      if (( dup = str_dup( item->str, item->used )) == NULL )
         exit( 1 );

      STACK_PUSH( type_table, dup )
      if ( alloc_err )
         exit( 1 );

      item->free += item->used;
      item->used = 0;
      item->top = item->str;
   }

   close( fd );
   string_free( item );

   if ( ! type_table->used || type_table->used % 2 )
   {
      fprintf( stderr, "drood: empty field in %s\n", TYPEFILE );
      exit( 1 );
   }

   STACK_PUSH( type_table, "" )
   if ( alloc_err )
      exit( 1 );
}

void add_scgi_server( char *optarg )
{
   char *host, *port, *suffix;

   /*
    * Need to work on copy or strsep() will insert nuls into the argv
    * string to which optarg is pointing.  This will screw up
    * restart_drood() which needs the original arguments.
    */

   if (( optarg = str_dup( optarg, strlen( optarg ))) == NULL )
      exit( 1 );

   host = strsep( &optarg, "|" );
   port = strsep( &optarg, "|" );
   suffix = optarg;

   if ( host == NULL || port == NULL || suffix == NULL )
   {
      fputs( "drood: incomplete -s option value\n", stderr );
      exit( 1 );
   }

   if ( scgi_servers == NULL && ( scgi_servers = make_stack()) == NULL )
      exit( 1 );

   STACK_PUSH( scgi_servers, suffix )
   if ( alloc_err )
      exit( 1 );

   STACK_PUSH( scgi_servers, host )
   if ( alloc_err )
      exit( 1 );

   STACK_PUSH( scgi_servers, port )
   if ( alloc_err )
      exit( 1 );

   if ( alloc_err )
      exit( 1 );

   scgi = 1;
}

void set_options( int argc, char **argv )
{
   int i;
   char **ptr;
   static char buffer[ PATH_MAX ];

   /*
    * Create a global version of argv for use by restart_drood().  That
    * function uses execv() to avoid searching PATH, so the first arg
    * is replaced by its expansion by realpath().  All the succeeding
    * argument pointers are copied from argv.
    */
   
   if (( g_argv = memory(( argc + 1 ) * sizeof( char * ))) == NULL )
      exit( 1 );

   if (( *g_argv = realpath( *argv, buffer )) == NULL )
      exit( 1 );

   for( i = 1, ptr = &argv[ 1 ]; i < argc; ++i, ++ptr )
      g_argv[ i ] = *ptr;

   g_argv[ i ] = NULL;

   /*
    * Process the args.
    */

   while(( i = getopt( argc, argv, "xl:s:q:t:w:i:p:r:d:u:g:e:m:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'c':
            certfile = optarg;
            break;

         case 'e':
            root_resource = optarg;
            break;

         case 'g':
            grp = optarg;
            break;

         case 'i':
            interface = optarg;
            break;

         case 'k':
            keyfile = optarg;
            break;

         case 'm':
            instances = strtol( optarg, NULL, 10 );
            break;

         case 'l':
            logfile = optarg;
            break;

         case 'p':
            port = optarg;
            break;

         case 'r':
            root = optarg;
            break;

         case 's':
            add_scgi_server( optarg );
            break;

         case 't':
            read_timeout = strtol( optarg, NULL, 10 );
            break;

         case 'd':
            dynamic = optarg;
            break;

         case 'u':
            user = optarg;
            break;

         case 'q':
            backlog = strtol( optarg, NULL, 10 );
            break;

         case 'w':
            write_timeout = strtol( optarg, NULL, 10 );
            break;

         case 'x':
            testing = 1;
      }

   if ( root == NULL )
   {
      fputs( "drood: server root directory not specified\n", stderr );
      exit( 1 );
   }

   if ( instances < 1 )
   {
      int mib[ 4 ];
      size_t len;

      len = sizeof( instances );

      mib[ 0 ] = CTL_HW;
      mib[ 1 ] = HW_NCPU;

      if ( sysctl( mib, 2, &instances, &len, NULL, 0 ) < 0 || instances < 0 )
      {
         fprintf( stderr, "sysctl(): %s\n", strerror( errno ));
         fputs( "Unable to determine number of CPUs.\n"
                "Running 1 instance of Drood.\n"
                "Use the -m option if this is not satisfactory.\n", stderr );

         instances = 1;
      }
   }

   if ( dynamic != NULL )
   {
      struct string *s;
      char *ptr;

      clean_dynamic = dynamic;

      if ( *dynamic == '.' &&
           ( *( dynamic + 1 ) == 0 || ( *( dynamic + 1 ) == '.' && *( dynamic + 2 ) == 0 )))
      {
         fprintf( stderr,
                  "drood: CGI directory name cannot be \".\" or \"..\"\n" );
         exit( 1 );
      }

      if (( s = make_string() ) == NULL )
         exit( 1 );

      STRING_APPEND( s, '/' )
      if ( alloc_err )
         exit( 1 );

      for( ptr = dynamic; *ptr; ++ptr )
         if ( *ptr == '/' )
         {
            fprintf( stderr,
                     "drood: CGI directory name contains virgule(s)\n" );
            exit( 1 );
         }
         else
         {
            STRING_APPEND( s, *ptr )
            if ( alloc_err )
               exit( 1 );
         }

      STRING_APPEND( s, '/' )
      if ( alloc_err )
         exit( 1 );

      dynamic = s->str;
      dynalen = s->used;

      free( s ); /* not string_free() */
   }

   if ( read_timeout < 1 || read_timeout > 60 )
   {
      fprintf( stderr, "drood: read timeout value %d out of range 1-60\n",
               read_timeout );
      exit( 1 );
   }

   if ( write_timeout < 0 || write_timeout > 120 )
   {
      fprintf( stderr, "drood: write timeout value %d out of range 0-120\n",
               write_timeout );
      exit( 1 );
   }

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "drood: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "drood: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }

   root_resource_len = strlen( root_resource );

   if ( dynamic == NULL )
      errstrings[ 5 ] = " 405 Method Not Allowed\r\nAllow: GET, HEAD\r\n";
}

void reset_signals()
{
   int *iptr;
   int sigs[] = { SIGTERM, SIGHUP, SIGPIPE, SIGQUIT, SIGUSR1, SIGUSR2,
                  SIGALRM, SIGINT, SIGTSTP, -1 };

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, SIG_DFL );
}

void set_signals()
{
   struct sigaction sa;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * Prevents zombies.
    */

   signal( SIGCHLD, SIG_IGN );

   /*
    * When the master process receives a SIGTERM, it propagates it to its
    * server children.  Master and servers exit as a consquence.  Here we
    * set SIGTERM to interrupt system calls, so that kevent() will return
    * immediately.
    */

   sa.sa_handler = signal_handler;
   sa.sa_flags = 0;
   sigaction( SIGTERM, &sa, NULL );

   /*
    * When not testing, master and server processes ignore signals in list.
    * When testing, all signals in list cause lone server process to exit.

    * When testing, there is only one server process and no master process.
    */

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( testing ? signal_handler : SIG_IGN ));
}

void become_daemon()
{
   int file, count, restart = 0;
   char buffer[ 16 ], buffer2[ 16 ];

   umask( 0 );
   snprintf( buffer2, sizeof( buffer2 ), "%d", getpid() );

   if (( file = open( PIDFILE, O_RDONLY )) > 0 )
   {
      if (( count = read( file, buffer, sizeof( buffer ) - 1 )) > 0 )
      {
         buffer[ count ] = '\0';

         /*
          * If the current pid is the same as the pid in PIDFILE,
          * then we are restarting in response to a SIGHUP.
          */

         if ( !strcmp( buffer, buffer2 ))
            restart = 1;
      }

      close( file );

      if ( restart )
      {
         syslog( LOG_ERR, "SIGHUP received.  Restarting..." );
         return;
      }
   }

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "drood: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Our pid has changed after the fork() above.
    */

   snprintf( buffer2, sizeof( buffer2 ), "%d", getpid() );

   if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC,
                      S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
      syslog( LOG_WARNING, "open( %s ): %m", PIDFILE );
   else
   {
      write( file, buffer2, strlen( buffer2 ));
      close( file );
   }
}

void change_identity()
{
   if ( setgid( group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "setgid(): %m" );
      exit( 1 );
   }

   if ( setuid( passwd->pw_uid ) < 0 )
   {
      syslog( LOG_ERR, "setuid(): %m" );
      exit( 1 );
   }
}

void start_listening()
{
   struct addrinfo hints, *res;
   struct accept_filter_arg afa;
   int result;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified a listening interface, we use
    * the IPv6 wildcard.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   result = 0;

   /*
    * Note we are shutting off this option if we are binding to the IPv6
    * wildcard interface, so we may accept both IPv4 and IPv6 connections.
    */

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   /*
    * Accept filter must be installed after socket is listening.
    */

   bzero( &afa, sizeof( afa ));
   strcpy( afa.af_name, "httpready" );

   if ( setsockopt( fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa,
                    sizeof( afa )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_ACCEPTFILTER ): %m" );

   close_on_exec( fd );
   non_blocking( fd );
}

void add_conn( int new, char *addr )
{
   struct ccb *ptr;
   char *aptr;
   int len;

   /*
    * Drop the connection if we have run out of memory.
    */

   if (( ptr = memory( sizeof( struct ccb ))) == NULL )
   {
      close( new );
      return;
   }

   for( len = 0, aptr = addr; *aptr; ++aptr )
      ++len;

   if (( ptr->addr = str_dup( addr, len )) == NULL )
   {
      free( ptr );
      close( new );
      return;
   }

   ptr->working = NULL;
   *ptr->buffer = '\0';
   *ptr->inbuf = '\0';
   ptr->inptr = ptr->inbuf;

   ptr->close = ptr->expect = ptr->count = ptr->error = ptr->inactive =
      ptr->gzip = ptr->timer = 0;

   ptr->first = ( dynamic != NULL ? 1 : 0 );

   ptr->size = ptr->written = ptr->scgi_len = 0;

   ptr->state = RECEIVE;
   ptr->next = NULL;

   ptr->source = ptr->verno = ptr->type = -1;
   ptr->sock = new;

   ptr->host = ptr->length = ptr->version = ptr->agent = ptr->cookie =
      ptr->request = ptr->method = ptr->since = ptr->unmodified =
      ptr->range = ptr->referer = ptr->content = ptr->scgi_host =
      ptr->scgi_port = NULL;

   ptr->request_found = ptr->close_found = ptr->expect_found = ptr->gzip_found =
      ptr->host_found = ptr->type_found = ptr->agent_found = ptr->length_found =
      ptr->since_found = ptr->unmodified_found = ptr->referrer_found =
      ptr->date_found = ptr->location_found = ptr->range_found = 0;

   ev_set( ptr->sock, EVFILT_READ, EV_ADD, ptr );
   ev_set( ptr->sock, EVFILT_WRITE, EV_ADD | EV_DISABLE, ptr );
   set_timer( ptr, 1 );
}

void clear_conn( struct ccb *item )
{
   char *ptr;

   if ( item->source >= 0 )
   {
      close( item->source );
      item->source = -1;
   }

   if ( item->host != NULL )
   {
      free( item->host );
      item->host = NULL;
   }

   if ( item->length != NULL )
   {
      free( item->length );
      item->length = NULL;
   }

   if ( item->version != NULL )
   {
      free( item->version );
      item->version = NULL;
   }

   if ( item->agent != NULL )
   {
      free( item->agent );
      item->agent = NULL;
   }

   if ( item->cookie != NULL )
   {
      free( item->cookie );
      item->cookie = NULL;
   }

   if ( item->request != NULL )
   {
      free( item->request );
      item->request = NULL;
   }

   if ( item->method != NULL )
   {
      free( item->method );
      item->method = NULL;
   }

   if ( item->since != NULL )
   {
      free( item->since );
      item->since = NULL;
   }

   if ( item->unmodified != NULL )
   {
      free( item->unmodified );
      item->unmodified = NULL;
   }

   if ( item->referer != NULL )
   {
      free( item->referer );
      item->referer = NULL;
   }

   if ( item->content != NULL )
   {
      free( item->content );
      item->content = NULL;
   }

   if ( item->range != NULL )
   {
      free( item->range );
      item->range = NULL;
   }

   if ( item->scgi_host != NULL )
   {
      /*
       * Do not free!  Pointing to element in scgi_servers.
      */

      item->scgi_host = NULL;
   }

   if ( item->scgi_port != NULL )
   {
      /*
       * Do not free!  Pointing to element in scgi_servers.
       */

      item->scgi_port = NULL;
   }

   if ( item->working != NULL )
   {
      while(( ptr = ( char *)stack_pop( item->working )) != NULL )
         free( ptr );
   }

   item->first = ( dynamic != NULL ? 1 : 0 );

   *item->buffer = '\0';
   item->close = item->count = item->error = item->expect = item->gzip =
      item->inactive = item->timer = 0;

   item->request_found = item->close_found = item->expect_found =
      item->host_found = item->type_found = item->gzip_found =
      item->agent_found = item->length_found = item->since_found =
      item->unmodified_found = item->referrer_found = item->range_found =
      item->date_found = item->location_found = 0;

   item->state = RECEIVE;
   item->next = NULL;

   item->size = item->written = item->scgi_len = 0;
   item->type = item->verno = -1;

   if ( item->sock >= 0 )
   {
      ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );

      /*
       * If there's still data in the input buffer, we have another
       * request to service.  The whole request may be in the buffer,
       * so we cannot rely on the socket becoming readable again to trigger
       * receive_request().

       * Otherwise, if the input buffer is empty, we shut off write
       * events for the socket, and we wait for read events to trigger
       * receive_request().
       */

      if ( *item->inptr )
         ev_set( item->sock, EVFILT_WRITE, EV_ENABLE, item );
      else
         ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );

      delete_timer( item );
      set_timer( item, 1 );
   }
}

void remove_conn( struct ccb *item )
{
   if ( item->sock < 0 )
      return;

   if ( item->addr != NULL )
      free( item->addr );

   closed = item->sock;
   delete_timer( item );

   close( item->sock );
   item->sock = -1;

   clear_conn( item );

   if ( item->working != NULL )
      stack_free( item->working );

   free( item );
   --active;
}

char **extract_request( char *string )
{
   static char *results[ 3 ];
   char *ptr, *ptr2, *buffer;
   int count;

   for( count = 0, ptr = string; *ptr; ++ptr )
      ++count;

   if (( buffer = memory( count )) == NULL )
      return NULL;

   alloc_err = 1;
   results[ 0 ] = results[ 1 ] = results[ 2 ] = NULL;
   count = 0;

   for( ptr2 = buffer, ptr = string; *ptr && *ptr != ' ' && *ptr != '\t'; )
   {
      ++count;
      *ptr2++ = *ptr++;
   }

   *ptr2 = '\0';

   if (( results[ 0 ] = str_dup( buffer, count )) == NULL )
   {
      free( buffer );
      return NULL;
   }

   while( *ptr == ' ' || *ptr == '\t' )
      ++ptr;

   ptr2 = buffer;
   count = 0;

   while( *ptr && *ptr != ' ' && *ptr != '\t' )
   {
      ++count;
      *ptr2++ = *ptr++;
   }

   *ptr2 = '\0';

   if (( results[ 1 ] = str_dup( buffer, count )) == NULL )
   {
      free( results[ 0 ] );
      free( buffer );
      return NULL;
   }

   while( *ptr == ' ' || *ptr == '\t' )
      ++ptr;

   ptr2 = buffer;
   count = 0;

   while( *ptr && *ptr != ' ' && *ptr != '\t' )
   {
      ++count;
      *ptr2++ = *ptr++;
   }

   *ptr2 = '\0';
   
   if (( results[ 2 ] = str_dup( buffer, count )) == NULL )
   {
      free( results[ 0 ] );
      free( results[ 1 ] );
      free( buffer );
      return NULL;
   }

   if ( ! *results[ 0 ] || ! *results[ 1 ] || ! *results [ 2 ] )
   {
      free( results[ 0 ] );
      free( results[ 1 ] );
      free( results[ 2 ] );

      alloc_err = 0;
      free( buffer );
      return NULL;
   }

   alloc_err = 0;
   free( buffer );
   return results;
}

char **extract_header( char *pattern, char *string )
{
   char *buffer, *ptr;
   static char *result = NULL;
   char *ptr1, *ptr2;
   int count;

   alloc_err = 0;

   ptr1 = pattern;
   ptr2 = string;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( *ptr1 )
      return NULL;

   while( *ptr2 && ( *ptr2 == ' ' || *ptr2 == '\t' ))
      ++ptr2;

   for( count = 0, ptr = string; *ptr; ++ptr )
      ++count;

   if (( buffer = memory( count )) == NULL )
   {
      alloc_err = 1;
      return NULL;
   }

   ptr = buffer;
   count = 0;

   while( *ptr2 && ( *ptr2 != ' ' && *ptr2 != '\t' ))
   {
      ++count;
      *ptr++ = *ptr2++;
   }

   if ( ! count )
   {
      free( buffer );
      return NULL;
   }

   if (( result = str_dup( buffer, count )) == NULL )
   {
      alloc_err = 1;
      free( buffer );
      return NULL;
   }

   free( buffer );
   return &result;
}

char **extract_content_type( char *string )
{
   char *buffer, *ptr;
   static char *result = NULL;
   char *ptr1, *ptr2;
   int count;

   alloc_err = 0;

   ptr1 = "content-type:";
   ptr2 = string;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( *ptr1 )
      return NULL;

   while( *ptr2 && ( *ptr2 == ' ' || *ptr2 == '\t' ))
      ++ptr2;

   for( count = 0, ptr = string; *ptr; ++ptr )
      ++count;
   
   if (( buffer = memory( count )) == NULL )
   {
      alloc_err = 1;
      return NULL;
   }

   ptr = buffer;
   count = 0;

   while( *ptr2 )
   {
      ++count;
      *ptr++ = *ptr2++;
   }

   if ( ! count )
   {
      free( buffer );
      return NULL;
   }

   if (( result = str_dup( buffer, count )) == NULL )
   {
      alloc_err = 1;
      free( buffer );
      return NULL;
   }

   free( buffer );
   return &result;
}

int find_gzip( struct ccb *item, char *line )
{
   char *ptr1, *ptr2, *saved, **header;

   if (( header = extract_header( "accept-encoding:", line )) == NULL )
      return 0;

   /*
    * The client accepts all encodings.
    */

   line = *header;

   if ( *line == '*' )
   {
      item->gzip = 1;
      free( *header );
      return 1;
   }

   ptr2 = line;

   /*
    * Otherwise find an occurrence of "gzip".
    */

AGAIN:
   do
   {
      ptr1 = "gzip";

      while( *ptr2 && *ptr2++ != *ptr1 )
         ;

      if ( ! *ptr2-- )
      {
         free( *header );
         return 0;
      }

      while( *ptr1 && *ptr2 && *ptr2 == *ptr1 )
      {
         ++ptr1;
         ++ptr2;
      }
   }
   while( *ptr1 );

   /*
    * If ptr1 has not been consumed, gzip does not occur on the line.
    */

   if ( *ptr1 )
   {
      free( *header );
      return 0;
   }

   /*
    * If this occurrence of "gzip" is the suffix of a longer word, we look
    * for another occurrence.
    */

   if (( ptr2 > line + 4 ) &&
        *( ptr2 - 5 ) != ' ' && *( ptr2 - 5 ) != '\t' && *( ptr2 - 5 ) != ',' )
      goto AGAIN;

   /*
    * If ptr1 has been consumed, and ptr2 has been consumed or is at a
    * comma, we have found "gzip" by itself without a qvalue.  We
    * interpret a space after "gzip" as a comma.
    */

   if ( ! *ptr2 || *ptr2 == ',' || *ptr2 == ' ' )
   {
      item->gzip = 1;
      free( *header );
      return 1;
   }

   /*
    * Otherwise, look for a qvalue in the unconsumed portion of ptr2.
    */

   ptr1 = ";q=";
   saved = ptr2;

TWICE:
   while( *ptr1 && *ptr2 && *ptr2 == *ptr1 )
   {
      ++ptr1;
      ++ptr2;
   }

   /*
    * If not found, perhaps there was a space before the qvalue.
    */

   if ( *ptr1 )
   {
      if ( saved != NULL )
      {
         ptr1 = "; q=";
         ptr2 = saved;
         saved = NULL;
         goto TWICE;
      }

      /*
       * If still not found, perhaps "gzip" was the prefix of another word,
       * like "gzipper", so we look for another occurrence of "gzip".
       */

      goto AGAIN;
   }

   /*
    * We have a qvalue.  If ptr2 was consumed, or the qvalue is zero, the
    * client cannot accept gzip encoding.
    */

   if ( *ptr2 && *ptr2 == '0' )
   {
      if ( *( ptr2 + 1 ) != '.' )
      {
         free( *header );
         return 0;
      }
      else if ( *( ptr2 + 2 ) == '0' )
      {
         free( *header );
         return 0;
      }
   }

   free( *header );
   item->gzip = 1;
   return 1;
}

void receive_request( struct ccb *item )
{
   char *line, **matches;
   int r, notready;

   for( ; ; )
   {
      notready = 0;

      if ( ! item->inactive )
      {
         item->count = 0;
         *item->buffer = '\0';
      }

      for( ; ; )
      {
         if ( !*item->inptr )
         {
            r = read( item->sock, item->inbuf, ( item->first ? 1 : INBUFSIZE - 1 ));

            if ( r <= 0 )
            {
               if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
                  notready = 1;

               line = NULL;
               break;
            }

            item->inbuf[ r ] = '\0';
            item->inptr = item->inbuf;
         }

         /*
          * We don't want the line terminators.
          */

         if ( *item->inptr == 13 )
         {
            ++item->inptr;
            continue;
         }

         if ( *item->inptr == 10 )
         {
            ++item->inptr;
            line = item->buffer;
            break;
         }

         /*
          * New character overwrites previous string terminator.  New string
          * terminator added.  Notice that item->count does not count the
          * string terminator.
          */

         item->buffer[ item->count++ ] = *item->inptr++;

         if ( item->count >= IOBUFSIZE )
         {
            notready = 2;
            line = NULL;
            break;
         }

         item->buffer[ item->count ] = '\0';
      }

      if ( line == NULL )
      {
         switch( notready )
         {
            /*
             * 0 means premature EOF.
             */

            case 0:
               remove_conn( item );
               break;

            /*
             * 1 means the connection has become inactive.
             */

            case 1:
               item->inactive = 1;
               break;

            /*
             * 2 means the client sent a too-long line.  If this happens,
             * there is most likely malicious intent, and the request header
             * is now incomplete, so we drop the connection after sending
             * the error message.
             */

            case 2:
               syslog( LOG_ERR, "overlong request header line from %s", item->addr );
               item->count = 0;
               item->error = ERR_BADREQUEST;
               item->state = ERROR;

               delete_timer( item );
               flip_events( item );
         }

         return;
      }

      item->inactive = 0;

      /*
       * If we have the request header separator line, we are done here.
       */

      if ( *line == '\0' )
      {
         item->count = 0;
         delete_timer( item );
         flip_events( item );

         if ( item->method == NULL || item->request == NULL || item->version == NULL )
         {
            syslog( LOG_ERR, "missing method, request, and/or HTTP version in request line from %s", item->addr );
            item->state = ERROR;
            item->error = ERR_BADREQUEST;
         }
         else
            analyze_request( item );

         return;
      }

      if ( !item->request_found )
      {
         if (( matches = extract_request( line )) == NULL )
         {
            if ( alloc_err )
            {
               remove_conn( item );
               return;
            }

            syslog( LOG_ERR, "missing or malformed request line from %s", item->addr );
            item->state = ERROR;
            item->error = ERR_BADREQUEST;

            delete_timer( item );
            flip_events( item );
            return;
         }
         else
         {
            item->request_found = 1;
            item->method = matches[ 0 ];

            if ( dynamic != NULL )
            {
               char *ptr1, *ptr2;

               ptr1 = item->method;
               ptr2 = "POST";

               while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
               {
                  ++ptr1;
                  ++ptr2;
               }

               if ( *ptr2 )
                  item->first = 0;
            }

            item->request = matches[ 1 ];
            item->version = matches[ 2 ];
            continue;
         }
      }
      else
      {
         char *ptr1;

         /*
          * Convert header keyword to lowercase.
          */

         for( ptr1 = line; *ptr1; ++ptr1 )
            if ( *ptr1 == ':' )
               break;
            else if ( *ptr1 >= 65 && *ptr1 <= 90 )
               *ptr1 += 32;
      }

      if ( *line == 'c' )
      {
         if ( !item->close_found )
         {
            char *ptr1, *ptr2;

            ptr1 = line;
            ptr2 = "connection: close";

            while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
            {
               ++ptr1;
               ++ptr2;
            }

            if ( ! *ptr2 )
            {
               item->close_found = 1;
               item->close = 1;
               continue;
            }
         }

         if ( !item->type_found )
         {
            if (( matches = extract_content_type( line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->type_found = 1;
               item->content = matches[ 0 ];
               continue;
            }
         }

         if ( !item->length_found )
         {
            if (( matches = extract_header( "content-length:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->length_found = 1;
               item->length = matches[ 0 ];
               continue;
            }
         }

         /*
          * We do not set a flag for the Cookie header, since there can be
          * more than one, legally.
          */

         if (( matches = extract_header( "cookie:", line )) == NULL )
         {
            if ( alloc_err )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            /*
             * Combine multiple cookies.
             */

            if ( item->cookie != NULL )
            {
               int len, len2, i;
               char *ptr, *ptr2;

               for( len = 0, ptr = item->cookie; *ptr; ++ptr )
                  ++len;

               for( len2 = 0, ptr = matches[ 0 ]; *ptr; ++ptr )
                  ++len2;

               if (( line = memory( len + len2 + 3 )) == NULL )
               {
                  remove_conn( item );
                  return;
               }

               ptr = line;
               ptr2 = item->cookie;

               for( i = len; i; --i )
                  *ptr++ = *ptr2++;

               line[ len ] = ';';
               line[ len + 1 ] = ' ';

               ptr = &line[ len + 2 ];
               ptr2 = matches[ 0 ];

               for( i = len2; i; --i )
                  *ptr++ = *ptr2++;

               line[ len + len2 + 2 ] = '\0';

               free( item->cookie );
               item->cookie = line;
            }
            else
               item->cookie = matches[ 0 ];
         }

         continue;
      }

      if ( *line == 'h' )
      {
         if ( !item->host_found )
         {
            if (( matches = extract_header( "host:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->host_found = 1;
               item->host = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'u' )
      {
         if ( !item->agent_found )
         {
            if (( matches = extract_header( "user-agent:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->agent_found = 1;
               item->agent = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'i' )
      {
         if ( !item->since_found )
         {
            if (( matches =  extract_header( "if-modified-since:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->since_found = 1;
               item->since = matches[ 0 ];
               continue;
            }
         }

         if ( !item->unmodified_found )
         {
            if (( matches =  extract_header( "if-unmodified-since:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->unmodified_found = 1;
               item->unmodified = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'e' )
      {
         if ( !item->expect_found )
         {
            char *ptr1, *ptr2;

            ptr1 = line;
            ptr2 = "expect: 100-continue";

            while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
            {
               ++ptr1;
               ++ptr2;
            }

            if ( ! *ptr2 )
            {
               item->expect_found = 1;
               item->expect = 1;
            }
         }

         continue;
      }

      if ( *line == 'r' )
      {
         if ( !item->range_found )
         {
            if (( matches = extract_header( "range:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->range_found = 1;
               item->range = matches[ 0 ];
               continue;
            }
         }

         if ( !item->referrer_found )
         {
            if (( matches = extract_header( "referer:", line )) == NULL )
            {
               if ( alloc_err )
               {
                  remove_conn( item );
                  return;
               }
            }
            else
            {
               item->referrer_found = 1;
               item->referer = matches[ 0 ];
            }
         }

         continue;
      }

      if ( *line == 'a' )
      {
         if ( ! item->gzip_found && item->count > 7 && line[ 7 ] == 'e' )
         {
            if ( !( item->gzip_found = find_gzip( item, line )) && alloc_err )
            {
               remove_conn( item );
               return;
            }
         }
      }
   }
}

char *date_header()
{
   static char buffer[ 64 ];
   static time_t t = 0, ot = 0;
   struct tm *lt;
   int len;

   if (( t = time( NULL )) != ot )
   {
      ot = t;
      lt = gmtime( &t );
      len = strftime( buffer, sizeof( buffer ),
                      "Date: %a, %d %b %Y %H:%M:%S %Z", lt );

      buffer[ --len ] = 'T';
      buffer[ --len ] = 'M';
      buffer[ --len ] = 'G';

      strcat( buffer, "\r\n" );
   }

   return buffer;
}

int fork_cgi( char *prog, struct ccb *item )
{
   int pfd[ 2 ], flags, pid, ourend;
   char *ptr, *ptr2, *args[ 2 ];

   ourend = -1;

   if ( pipe( pfd ) < 0 )
   {
      syslog( LOG_ERR, "pipe(): %m" );
      return ourend;
   }

   switch(( pid = fork() ))
   {
      case -1:
         close( pfd[ 0 ] );
         close( pfd[ 1 ] );
         syslog( LOG_ERR, "fork(): %m" );
         return ourend;

      case 0:
         change_identity();

         reset_signals();
         fclose( stdout );

         if (( dup2( pfd[ 1 ], 1 )) < 0 )
         {
            syslog( LOG_ERR, "dup2: %m" );
            _exit( 1 );
         }

         close( pfd[ 0 ] );
         close( pfd[ 1 ] );

         if ( item->type == 2 )
         {
            fclose( stdin );

            if (( dup2( item->sock, 0 )) < 0 )
            {
               syslog( LOG_ERR, "dup2: %m" );
               _exit( 1 );
            }

            if (( flags = fcntl( 0, F_GETFL, 0 )) < 0 )
            {
               syslog( LOG_ERR, "fcntl(): %m" );
               _exit( 1 );
            }

            if ( fcntl( 0, F_SETFL, flags & ~O_NONBLOCK ) < 0 )
            {
               syslog( LOG_ERR, "fcntl(): %m" );
               _exit( 1 );
            }
         }

         close( item->sock );

         while(( ptr = stack_pop( item->working )) != NULL )
         {
            ptr2 = stack_pop( item->working );
            setenv( ptr2, ptr, 1 );
            free( ptr );
            free( ptr2 );
         }

         setenv( "SCGI", "0", 1 );
         args[ 0 ] = prog;
         args[ 1 ] = NULL;

         closelog();
         execv( args[ 0 ], args );
         syslog( LOG_ERR, "execv(): %m" );
         _exit( 1 );

      default:
         close( pfd[ 1 ] );
         ourend = pfd[ 0 ];
         close_on_exec( ourend );
         non_blocking( ourend );
   }

   return ourend;
}

void send_scgi_body( struct ccb *item )
{
   off_t written;

   if ( ! item->count )
   {
      item->count = read( item->sock, item->buffer,
                          MIN( item->scgi_len, sizeof( item->buffer )));

      if ( item->count < 0 )
      {
         remove_conn( item );
         return;
      }

      /*
       * Since we do not read more than item->scgi_len chars, this
       * can only happen if the client closed (at least) the readable
       * half of the connection, prematurely.  This means the request
       * is borked, so we bail.
       */

      if ( ! item->count )
      {
         remove_conn( item );
         return;
      }

      item->written = 0;
   }

   written = write( item->source, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;
   item->scgi_len -= written;

   if ( ! item->scgi_len )
   {
      *item->buffer = '\0';

      ev_set( item->source, EVFILT_READ, EV_ADD, item );
      ev_set( item->sock, EVFILT_READ, EV_DISABLE, item );

      item->type_found = item->date_found = item->location_found =item->length_found = 0;
         item->length_found = 0;
      item->size = 0;
      item->state = FILTER;
   }
}

void send_scgi_header( struct ccb *item )
{
   off_t written;

   written = write( item->source, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->written = 0;
      ev_set( item->source, EVFILT_WRITE, EV_DISABLE, item );

      if ( item->scgi_len )
      {
         item->state = SCGI_BODY;
         ev_set( item->sock, EVFILT_READ, EV_ENABLE, item );
      }
      else
      {
         *item->buffer = '\0';
         item->type_found = item->date_found = item->location_found =
            item->length_found = 0;

         item->size = 0;
         item->state = FILTER;
         ev_set( item->source, EVFILT_READ, EV_ADD | EV_ENABLE, item );
      }
   }
}

void prepare_scgi( struct ccb *item )
{
   char *var, *val, *ptr;
   int i, err, len, ns_len, total;
   socklen_t slen;

   slen = sizeof( err );
   err = 0;

   if ( getsockopt( item->source, SOL_SOCKET, SO_ERROR, &err, &slen ) < 0 )
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "getsockopt(): %m" );
      return;
   }

   if ( err )
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, "connect(): %s", strerror( err ));
      return;
   }

   for( len = 0, i = 0; i < item->working->used; ++i )
   {
      for( ptr = item->working->values[ i ]; *ptr; ++ptr )
         ++len;
      ++len;
   }

   /*
    * Add seven characters for the SCGI header, added later.
    */

   len += 7;
   ns_len = snprintf( item->buffer, sizeof( item->buffer ), "%d:", len );

   /*
    * The extra character is for the comma,
    * added to the buffer after the CGI environment netstring.
    */

   if ( len + ns_len + 1 > IOBUFSIZE )
   {
      syslog( LOG_ERR, "SCGI header is too big to process" );
      item->state = ERROR;
      item->error = ERR_BADREQUEST;
      return;
   }

   /*
    * CONTENT_LENGTH was last item pushed onto working stack, so it
    * will be the first item popped off.  It must come first in the header
    * according to the SCGI protocol.
    */

   item->scgi_len = ( off_t )atol( *( char **)item->working->top );
   len = ns_len;

   while(( val = stack_pop( item->working )) != NULL )
   {
      var = stack_pop( item->working );

      for( total = 0, ptr = var; *ptr; ++ptr )
         ++total;

      ++total;
      ptr = var;

      for( i = len; total; ++i, --total )
      {
         item->buffer[ i ] = *ptr++;
         ++len;
      }

      for( total = 0, ptr = val; *ptr; ++ptr )
         ++total;

      ++total;
      ptr = val;

      for( i = len; total; ++i, --total )
      {
         item->buffer[ i ] = *ptr++;
         ++len;
      }

      free( var );
      free( val );
   }

   ptr = "SCGI";
   for( i = 0; i < 5; ++i )
      item->buffer[ len++ ] = *ptr++;

   item->buffer[ len++ ] = '1';
   item->buffer[ len++ ] = '\0';
   item->buffer[ len++ ] = ',';

   item->count = len;
   item->written = 0;
   item->state = SCGI_HEADER;
}

void connect_scgi( struct ccb *item )
{
   int fd = -1, result;
   char buffer[ 255 ];
   struct addrinfo hints, *res, *ptr;

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_family = PF_UNSPEC;
   hints.ai_socktype = SOCK_STREAM;

   if (( result = getaddrinfo( item->scgi_host, item->scgi_port, &hints, &res )))
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      snprintf( buffer, sizeof( buffer ), "getaddrinfo(): %s",
                gai_strerror( result ));
      syslog( LOG_ERR, "%s", buffer );
      return;
   }

   for( ptr = res; ptr != NULL; ptr = ptr->ai_next )
   {
      result = 0;
      fd = socket( ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol );

      if ( fd == -1 )
         continue;

      non_blocking( fd );
      close_on_exec( fd );

      if (( result = connect( fd, ptr->ai_addr, ptr->ai_addrlen )) == -1 )
      {
         if ( errno != EINPROGRESS )
         {
            close( fd );
            fd = -1;
            continue;
         }
      }

      break;
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( fd == -1 )
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;

      syslog( LOG_ERR, ( result ? "connect(): %m" : "socket(): %m" ));
      return;
   }

   item->source = fd;
   item->state = SCGI_PREPARE;

   ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
   ev_set( item->source, EVFILT_WRITE, EV_ADD, item );
}

int add_to_env( struct ccb *item, char *str )
{
   char *ptr;
   int count;

   for( count = 0, ptr = str; *ptr; ++ptr )
      ++count;

   if (( ptr = str_dup( str, count )) == NULL )
      return 1;

   STACK_PUSH( item->working, ptr )
   if ( alloc_err )
   {
      free( ptr );
      return 1;
   }

   return 0;
}

int find_scgi_server( struct ccb *item, char *prog, int len )
{
   char *ptr, *ptr1, *ptr2;
   int i;

   ptr = &prog[ len ];

   while( ptr >= prog && *ptr != '.' )
      --ptr;

   if ( ptr >= prog && *ptr == '.' )
      ++ptr;
   else
      return 0;

   for( i = 0; i < scgi_servers->used; i += 3 )
   {
      ptr1 = scgi_servers->values[ i ];
      ptr2 = ptr;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if( ! *ptr1 && ! *ptr2 )
      {
         item->scgi_host = scgi_servers->values[ i + 1 ];
         item->scgi_port = scgi_servers->values[ i + 2 ];
         return 1;
      }
   }

   item->scgi_host = NULL;
   item->scgi_port = NULL;

   return 0;
}

void prepare_dynamic( struct ccb *item )
{
   char *ptr, *args;
   struct stat st;
   static struct string *s = NULL;

   if ( item->expect )
   {
      item->count = snprintf( item->buffer, sizeof( item->buffer ),
                              "%s 100 Continue\r\n\r\n", item->version );
      item->expect = 0;
      item->state = HEADER;
      item->next = CONTINUE;
      return;
   }

   /*
    * We push all the environment variable data onto the working stack
    * so we can install them in the child process, avoiding the memory
    * leak described in setenv(3) in the parent.  This drove me crazy for
    * two days.  It is necessary to keep previous versions of environment
    * variable values because so that any previously returned pointers to
    * them remain valid for the lifetime of the process.
    */

   if ( item->working == NULL && ( item->working = make_stack()) == NULL )
   {
      remove_conn( item );
      return;
   }

   if ( add_to_env( item, "REMOTE_ADDR" ) ||
        add_to_env( item, item->addr ))
   {
      remove_conn( item );
      return;
   }

   if ( s == NULL && ( s = make_string()) == NULL )
   {
      remove_conn( item );
      return;
   }
   else
   {
      s->free += s->used;
      s->used = 0;
      s->top = s->str;
   }

   if ( add_to_env( item, "REQUEST_METHOD" ) || add_to_env( item, item->method ))
   {
      remove_conn( item );
      return;
   }

   STRING_APPEND( s, '.' )
   if ( alloc_err )
   {
      remove_conn( item );
      return;
   }

   for( ptr = item->request; *ptr; ++ptr )
      if ( *ptr == '?' )
         break;
      else
      {
         STRING_APPEND( s, *ptr )
         if ( alloc_err )
         {
            remove_conn( item );
            return;
         }
      }

   args = ( *ptr == '?' ? ++ptr : "" );

   if ( stat( s->str, &st ) < 0 )
   {
      item->state = ERROR;

      if ( errno == ENOENT || errno == ENOTDIR )
      {
         item->error = ERR_NOTFOUND;
         return;
      }

      if ( errno == EACCES )
      {
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->error = ERR_INTERNAL;
      return;
   }

   /*
    * Check for file type and permissions.
    */

   if ( ! S_ISREG( st.st_mode ) ||
        (
           ! (( S_IRUSR & st.st_mode ) && passwd->pw_uid == st.st_uid ) &&
           ! (( S_IRGRP & st.st_mode ) && group->gr_gid == st.st_gid ) &&
           ! ( S_IROTH & st.st_mode )
        ) ||
        (
           ! (( S_IXUSR & st.st_mode ) && passwd->pw_uid == st.st_uid ) &&
           ! (( S_IXGRP & st.st_mode ) && group->gr_gid == st.st_gid ) &&
           ! ( S_IXOTH & st.st_mode )
        ))
   {
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   /*
    * We refuse to run the CGI program just to generate a header.
    * This response merely indicates the CGI resource exists and is
    * runnable.
    */

   if ( ! item->type )
   {
      item->count = snprintf( item->buffer, IOBUFSIZE, "%s 200 OK\r\n%s%sContent-Length: 0\r\n\r\n",
                              item->version, server_header, date_header() );
      item->state = HEADER;
      item->next = HEADER;
      return;
   }

   if ( add_to_env( item, "SCRIPT_NAME" ) ||
        add_to_env( item, &s->str[ 1 ] ) ||
        add_to_env( item, "QUERY_STRING" ) ||
        add_to_env( item, args ) ||
        add_to_env( item, "SERVER_SOFTWARE" ) ||
        add_to_env( item, "Drood/" VERSION ) ||
        add_to_env( item, "GATEWAY_INTERFACE" ) ||
        add_to_env( item, "CGI/1.1" ) ||
        add_to_env( item, "HTTP_USER_AGENT" ) ||
        add_to_env( item, ( item->agent == NULL ? "" : item->agent )) ||
        add_to_env( item, "HTTP_COOKIE" ) ||
        add_to_env( item, ( item->cookie == NULL ? "" : item->cookie )) ||
        add_to_env( item, "SERVER_NAME" ) ||
        add_to_env( item, item->host ) ||
        add_to_env( item, "SERVER_PORT" ) ||
        add_to_env( item, port ) ||
        add_to_env( item, "CONTENT_TYPE" ) ||
        add_to_env( item, ( item->content == NULL ? "" : item->content )) ||
        add_to_env( item, "GZIP" ) ||
        add_to_env( item, ( item->gzip ? "1" : "0" )))
   {
      remove_conn( item );
      return;
   }

   if ( item->type == 1 )
   {
      if ( add_to_env( item, "CONTENT_LENGTH" ) || add_to_env( item, "0" ))
      {
         remove_conn( item );
         return;
      }
   }
   else
   {
      if ( item->length == NULL )
      {
         item->state = ERROR;
         item->error = ERR_LENGTHREQ;
         return;
      }

      if ( add_to_env( item, "CONTENT_LENGTH" ) ||
           add_to_env( item, item->length ))
      {
         remove_conn( item );
         return;
      }
   }

   /*
    * Will be re-used in filter_cgi_header().
    */

   if ( item->length != NULL )
   {
      free( item->length );
      item->length = NULL;
   }

   *item->buffer = '\0';
   item->count = 0;

   if ( scgi && find_scgi_server( item, s->str, s->used ))
   {
      item->state = SCGI_CONNECT;
      return;
   }
   else if (( item->source = fork_cgi( s->str, item )) < 0 )
   {
      item->state = ERROR;
      item->error = ERR_INTERNAL;
   }
   else
   {
      item->type_found = item->date_found = item->location_found =
         item->length_found = 0;

      item->state = FILTER;
      ev_set( item->source, EVFILT_READ, EV_ADD, item );
      ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
   }

   while(( ptr = stack_pop( item->working )) != NULL )
      free( ptr );
}

void prepare_static( struct ccb *item )
{
   static struct string *s = NULL;
   int len;
   struct stat st;
   struct tm *lt;
   char *ptr, **tp, type[ 128 ], size[ 128 ], modified[ 128 ];
   struct sc *sc_ptr;
   time_t t;

   if ( item->type == 2 )
   {
      syslog( LOG_ERR, "POST request for static resource from %s", item->addr );
      item->error = ERR_BADMETHOD;
      item->state = ERROR;
      return;
   }

   if ( s == NULL && ( s = make_string()) == NULL )
   {
      remove_conn( item );
      return;
   }

   s->free += s->used;
   s->used = 0;
   s->top = s->str;

   /*
    * We make sure there is at least one separator between the various
    * items of the path we are going to create in s, because the
    * command-line arguments may not have included them, and the resource
    * specified in the request may have not begun with a slash.  Multiple
    * separators are tolerated by the filesystem-related system calls.
    */

   STRING_APPEND( s, '.' )
   if ( alloc_err )
   {
      remove_conn( item );
      return;
   }

   STRING_APPEND( s, '/' )
   if ( alloc_err )
   {
      remove_conn( item );
      return;
   }

   for ( ptr = item->host; *ptr; ++ptr )
   {
      STRING_APPEND( s, *ptr )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }
   }

   STRING_APPEND( s, '/' )
   if ( alloc_err )
   {
      remove_conn( item );
      return;
   }

   for( ptr = item->request; *ptr; ++ptr )
   {
      STRING_APPEND( s, *ptr )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }
   }

AGAIN:
   if ( item->gzip && ! item->range_found )
   {
      STRING_APPEND( s, '.' )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }

      STRING_APPEND( s, 'g' )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }

      STRING_APPEND( s, 'z' )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }
   }

   t = time( NULL );

NO_GZIP:
   for( sc_ptr = stat_cache; sc_ptr->t; ++sc_ptr )
   {
      char *ptr1, *ptr2;

      if ( sc_ptr->file == NULL )
         continue;

      ptr1 = sc_ptr->file;
      ptr2 = s->str;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
      {
         if ( sc_ptr->t == t )
            st = sc_ptr->st;
         else
         {
            free( sc_ptr->file );
            bzero( sc_ptr, sizeof( struct sc ));
         }

         break;
      }
   }

   if ( ! sc_ptr->t && stat( s->str, &st ) < 0 )
   {
      if ( item->gzip && ! item->range_found )
      {
         s->top -= 3;
         s->free += 3;
         s->used -= 3;
         s->str[ s->used ] = '\0';

         item->gzip = 0;
         goto NO_GZIP;
      }

      if ( errno == ENOENT || errno == ENOTDIR )
      {
         item->state = ERROR;
         item->error = ERR_NOTFOUND;
         return;
      }

      if ( errno == EACCES )
      {
         item->state = ERROR;
         item->error = ERR_FORBIDDEN;
         return;
      }

      item->state = ERROR;
      item->error = ERR_INTERNAL;
      return;
   }

   if ( ! sc_ptr->t )
   {
      sc_ptr = &stat_cache[ CACHE_SIZE - 1 ];
      
      if ( sc_ptr->file != NULL )
         free( sc_ptr->file );
      
      for( ; sc_ptr > stat_cache; --sc_ptr )
         *sc_ptr = *( sc_ptr - 1 );
   
      stat_cache[ 0 ].t = t;
      stat_cache[ 0 ].file = str_dup( s->str, s->used );
      stat_cache[ 0 ].st = st;
   }
   
   if ( S_ISDIR( st.st_mode ))
   {
      char *ptr;

      STRING_APPEND( s, '/' )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }

      for( ptr = root_resource; *ptr; ++ptr )
      {
         STRING_APPEND( s, *ptr )
         if ( alloc_err )
         {
            remove_conn( item );
            return;
         }
      }

      goto AGAIN;
   }

   if ( ! S_ISREG( st.st_mode ))
   {
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   /*
    * Don't send any file which is executable.
    */

   if (( S_IXUSR & st.st_mode ) ||( S_IXGRP & st.st_mode ) ||
         S_IXOTH & st.st_mode )
   {
      item->state = ERROR;
      item->error = ERR_FORBIDDEN;
      return;
   }

   if ( ! st.st_size )
   {
      item->state = ERROR;
      item->error = ERR_NOCONTENT;
      return;
   }

   if ( item->type )
   {
      if (( item->source = open( s->str, O_RDONLY )) < 0 )
      {
         if ( errno == EACCES )
         {
            item->state = ERROR;
            item->error = ERR_FORBIDDEN;
            return;
         }

         item->state = ERROR;
         item->error = ERR_INTERNAL;

         syslog( LOG_ERR, "open(): %m" );
         return;
      }

      close_on_exec( item->source );
   }

   if ( item->gzip )
   {
      s->top -= 3;
      s->free += 3;
      s->used -= 3;
      s->str[ s->used ] = '\0';
   }

   ptr = &s->str[ s->used ];

   while( ptr >= s->str && *ptr != '.' )
      --ptr;

   if ( ptr >= s->str && *ptr == '.' )
      ++ptr;
   else
      ptr = NULL;

   if ( ptr == NULL )
      ptr = default_type;
   else
      for( tp = ( char **)type_table->values; **tp; tp += 2 )
      {
         char *ptr1, *ptr2;

         ptr1 = *tp;
         ptr2 = ptr;

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr1 && ! *ptr2 )
         {
            ptr = *( tp + 1 );
            break;
         }
      }

   lt = gmtime( &st.st_ctime );
   len = strftime( modified, sizeof( modified ),
                   "Last-Modified: %a, %d %b %Y %H:%M:%S %Z", lt );

   modified[ --len ] = 'T';
   modified[ --len ] = 'M';
   modified[ --len ] = 'G';

   if ( item->since != NULL )
   {
      char *ptr1, *ptr2;

      ptr1 = item->since;
      ptr2 = &modified[ 15 ];

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
      {
         item->state = ERROR;
         item->error = ERR_NOTMODIFIED;
         return;
      }
   }

   if ( item->unmodified != NULL )
   {
      char *ptr1, *ptr2;

      ptr1 = item->unmodified;
      ptr2 = &modified[ 15 ];

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( *ptr1 || *ptr2 )
      {
         item->state = ERROR;
         item->error = ERR_MODIFIED;
         return;
      }
   }

   strcat( modified, "\r\n" );
   snprintf( type, sizeof( type ), "Content-Type: %s\r\n", ptr );

   if ( item->type == 1 && item->range_found )
   {
      char *ptr1, *ptr2;
      off_t start, end;

      ptr1 = item->range;
      ptr2 = "bytes=";

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( *ptr2 )
      {
         item->state = ERROR;
         item->error = ERR_RANGE;
         return;
      }

      ptr2 = ptr1;

      while( *ptr2 && *ptr2 != '-' )
         ++ptr2;

      if ( ! *ptr2 )
      {
         item->state = ERROR;
         item->error = ERR_RANGE;
         return;
      }

      *ptr2++ = '\0';

      start = strtoll( ptr1, NULL, 10 );
      end = strtoll( ptr2, NULL, 10 );

      if ( start < 0 || end >= st.st_size || end < start )
      {
         item->state = ERROR;
         item->error = ERR_RANGE;
         return;
      }

      /*
       * Does double duty for scgi and range responses.
       */

      item->scgi_len = start;

      snprintf( size, sizeof( size ), "Content-Length: %llu\r\n",
                ( long long unsigned int)( item->size = end - start + 1 ));

      item->count = snprintf( item->buffer, IOBUFSIZE,
                              "%s 206 Partial Content\r\n%s%s%s%s%s"
                              "Content-Range: bytes %llu-%llu/%llu\r\n\r\n",
            
                              item->version, type, size, modified, server_header,
                              date_header(),
                              ( long long unsigned int)start,
                              ( long long unsigned int)end,
                              ( long long unsigned int)st.st_size );
   }
   else
   {
      snprintf( size, sizeof( size ), "Content-Length: %llu\r\n",
                ( item->type ? ( long long unsigned int )st.st_size :
                               ( long long unsigned int )0 ));

      item->count = snprintf( item->buffer, IOBUFSIZE, "%s 200 OK\r\n%s%s%s%s%s%s\r\n",
                              item->version, type, size,
                              ( item->gzip ? "Content-Encoding: gzip\r\n" : "" ),
                              modified, server_header, date_header() );
               
      item->size = st.st_size;
   }

   item->state = HEADER;
   item->next = ( item->type ? RESPOND : HEADER );
}

void analyze_request( struct ccb *item )
{
   static char *http = "HTTP/", *head = "HEAD", *get = "GET", *post = "POST";
   char *ptr1, *ptr2;
   int len;

   if ( logfd > 0 )
   {
      static time_t t = 0, ot = 0;
      static char buffer[ IOBUFSIZE ], datebuf[ 64 ];
      struct tm *lt;

      if (( t = time( NULL )) != ot )
      {
         ot = t;
         lt = gmtime( &t );
         strftime( datebuf, sizeof( datebuf ),
                   "%a, %d %b %Y %H:%M:%S %Z", lt );
      }

      len = snprintf( buffer, sizeof( buffer ), "%s|%s|%s|%s|http://%s/%s|%s\n",
                   datebuf, item->addr,
                   ( item->agent == NULL ? "no user-agent header" : item->agent ),
                   item->method,
                   ( item->host == NULL ? "default.host" : item->host ),
                   ( *item->request == '/' ? &item->request[ 1 ] : item->request ),
                   ( item->referer == NULL ? "no referrer header" : item->referer ));

      flock( logfd, LOCK_EX );

      if ( write( logfd, buffer, len ) < 0 )
         syslog( LOG_ERR, "write( logfd ): %m" );

      flock( logfd, LOCK_UN );
   }

   if ( item->host == NULL )
   {
      if (( item->host = str_dup( "default", 7 )) == NULL )
      {
         remove_conn( item );
         return;
      }
   }
   else if ( clean_dynamic != NULL )
   {
      ptr1 = item->host;
      ptr2 = clean_dynamic;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
      {
         syslog( LOG_ERR, "host header specifies dynamic resource directory from %s", item->addr );
         item->error = ERR_BADREQUEST;
         item->state = ERROR;
         return;
      }
   }

   ptr1 = item->method;
   ptr2 = head;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   item->type = -1;

   if ( ! *ptr1 && ! *ptr2 )
      item->type = 0;
   else
   {
      ptr1 = item->method;
      ptr2 = get;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         item->type = 1;
      else
      {
         ptr1 = item->method;
         ptr2 = post;

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr1 && ! *ptr2 )
            item->type = 2;
      }
   }

   if ( item->type < 0 )
   {
      item->error = ERR_BADMETHOD;
      item->state = ERROR;
      return;
   }

   ptr1 = item->version;
   ptr2 = http;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( *ptr2 )
   {
      syslog( LOG_ERR, "unrecognized HTTP version from %s", item->addr );
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   if ( ! *ptr1 || *ptr1 != '1'
        || ! *( ptr1 + 1 ) || *( ptr1 + 1 ) != '.'
        || ! *( ptr1 + 2 ) || ( *( ptr1 + 2 ) != '0' && *( ptr1 + 2 ) != '1' )
        || *( ptr1 + 3 ))
   {
      item->error = ERR_BADVERSION;
      item->state = ERROR;
      return;
   }

   if ( ! ( item->verno = ( *( ptr1 + 2 ) == '1' )))
      item->close = 1;

   /*
    * If the request is a POST, and we're not configured to run (S)CGI,
    * then we bail, closing the connection, to avoid having to read
    * past the request body to get to the next request header, if this is
    * an HTTP/1.1 connection.  Since errors of this sort are going to
    * be rare because they are most likely the result of a misconfigured
    * server, the simplest solution is just to force the client to open
    * another connection.
    */

   if ( item->type == 2 && dynamic == NULL )
   {
      item->error = ERR_BADMETHOD;
      item->state = ERROR;
      return;
   }

   if ( !( *item->request ))
   {
      syslog( LOG_ERR, "empty request from %s", item->addr );
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   /*
    * Reject requests for any path with /../ in it.
    */

   if ( *item->request == '.' || *item->host == '.' )
   {
      syslog( LOG_ERR, "request/host beginning with . from %s", item->addr );
      item->error = ERR_BADREQUEST;
      item->state = ERROR;
      return;
   }

   {
      char **ptr, *ptr2, *ptrs[ 3 ];

      ptrs[ 0 ] = item->host;
      ptrs[ 1 ] = item->request;
      ptrs[ 2 ] = NULL;

      for( ptr = ptrs; *ptr != NULL; ++ptr )
         for( ptr2 = *ptr; *ptr2; ++ptr2 )
            if ( *ptr2 && *ptr2 == '/'
                       && *( ptr2 + 1 ) && *( ptr2 + 1 )== '.'
                       && *( ptr2 + 2 ) && *( ptr2 + 2 ) == '.'
                       && ( ! *( ptr2 + 3 ) || *( ptr2 + 3 ) == '/' ))

            {
               syslog( LOG_ERR, "request/host with ../ in it from %s", item->addr );
               item->error = ERR_BADREQUEST;
               item->state = ERROR;
               return;
            }

      for( len = 0, ptr2 = item->request; *ptr2; ++ptr2 )
         ++len;
   }

   if ( item->request[ len - 1 ] == '/' )
   {
      char *ptr;

      len += root_resource_len;

      if (( ptr = memory( ++len )) == NULL )
      {
         remove_conn( item );
         return;
      }

      snprintf( ptr, len, "%s%s", item->request, root_resource );
      free( item->request );
      item->request = ptr;
   }

   if ( dynamic != NULL )
   {
      char *ptr1, *ptr2;

      ptr1 = dynamic;
      ptr2 = item->request;
      len = dynalen;

      while( len && *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
         --len;
      }

      if ( ! len )
      {
         prepare_dynamic( item );
         return;
      }
   }

   prepare_static( item );
}

/*
 * respond_header() and respond_error() are triggered by the writeability
 * of a client socket.  Because we know the socket is writable, we know
 * write() will not generate EWOULDBLOCK, but it can return short counts.
 * If write() returns an error, we drop the connection.
 */

void respond_header( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
   {
      item->written = 0;

      if (( item->state = item->next ) == HEADER )
         ( item->close ? remove_conn( item ) : clear_conn( item ) );
      else if ( item->state == CHUNK || item->state == NOCHUNK )
      {
         ev_set( item->source, EVFILT_READ, EV_ENABLE, item );
         ev_set( item->sock, EVFILT_WRITE, EV_DISABLE, item );
      }
      else if ( item->state != CONTINUE )
      {
         item->written = item->scgi_len;
         respond_static( item );
      }
   }
}

void respond_error( struct ccb *item )
{
   off_t written;

   if ( write_timeout && ! item->timer )
   {
      set_timer( item, 0 );
      item->timer = 1;
   }

   if ( ! item->count )
   {
      char buffer[ 512 ], len_buffer[ 64 ];
      int count;

      item->close = 1;

      count = snprintf( buffer, sizeof( buffer ),
                   "<html><head><title>Error</title></head><body>%s</body></html>",
                   errstrings[ item->error ] );

      snprintf( len_buffer, sizeof( len_buffer ), "Content-Length: %u\r\n", count );

      item->count = snprintf( item->buffer, IOBUFSIZE, "HTTP/1.0%sContent-Type: text/html\r\n"
                              "%s%s%sConnection: close\r\n\r\n<h1>%s</h1>",
                              errstrings[ item->error ], server_header, date_header(),
                              len_buffer, buffer );
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      remove_conn( item );
      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

void respond_static( struct ccb *item )
{
   off_t written = 0;

   if ( sendfile( item->source, item->sock, item->written,
                  MIN( item->size, IOBUFSIZE ), NULL,
                  &written, 0 ) < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EBUSY && errno != EAGAIN )
      {
         remove_conn( item );
         return;
      }
   }

   item->written += written;
   item->size -= written;

   if ( ! item->size )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

void respond_unchunked( struct ccb *item )
{
   int result;
   off_t written;

   if ( ! item->count )
   {
      switch(( result = read( item->source, item->buffer, IOBUFSIZE ) ))
      {
         case 0:
            ( item->close ? remove_conn( item ) : clear_conn( item ) );
            return;

         case -1:
            remove_conn( item );
            return;

         default:
            item->count = result;
            item->written = 0;
      }
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;
}

void respond_chunked( struct ccb *item )
{
   int result = 1;
   off_t written;

   if ( ! item->count )
   {
      if (( result = read( item->source, &item->buffer[ 6 ], IOBUFSIZE - 8 ))
            < 0 )
      {
         remove_conn( item );
         return;
      }

      item->buffer[ result + 6 ] = '\r';
      item->buffer[ result + 7 ] = '\n';

      item->count = result + 8;
      item->written = 0;

      snprintf( item->buffer, 6, "%.4X\r\n", result );
      item->buffer[ 5 ] = '\n';
   }

   written = write( item->sock, &item->buffer[ item->written ], item->count );

   if ( written < 0 )
   {
      if ( errno != EWOULDBLOCK && errno != EAGAIN )
         remove_conn( item );

      return;
   }

   item->count -= written;
   item->written += written;

   if ( ! item->count && ! result )
      ( item->close ? remove_conn( item ) : clear_conn( item ) );
}

void filter_cgi_header( struct ccb *item )
{
   char *line, *ptr, **matches, c;
   int r, notready;

   for( ; ; )
   {
      notready = 0;

      if ( ! item->inactive )
      {
         item->count = 0;
         *item->buffer = '\0';
      }

      for( ; ; )
      {
         r = read( item->source, &c, 1 );

         if ( r <= 0 )
         {
            if ( r < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ))
               notready = 1;

            line = NULL;
            break;
         }

         /*
          * We don't want the line terminators.
          */

         if ( c == 13 )
            continue;

         if ( c == 10 )
         {
            line = item->buffer;
            break;
         }

         /*
          * New character overwrites previous string terminator.  New string
          * terminator added.  Notice that item->count does not count the
          * string terminator.
          */

         item->buffer[ item->count++ ] = c;

         if ( item->count >= IOBUFSIZE )
         {
            notready = 2;
            line = NULL;
            break;
         }

         item->buffer[ item->count ] = '\0';
      }

      if ( line == NULL )
      {
         if ( notready == 1 )
         {
            item->inactive = 1;
            return;
         }
         else
         {
            /*
             * 0 means premature EOF from CGI program.
             * 2 means the CGI program sent a too-long response header line.
             * We bail in each case.
             */

            item->error = ERR_INTERNAL;
            item->state = ERROR;
            flip_events( item );
            return;
         }
      }

      /*
       * If we have the request header separator line, we are done here.
       */

      if ( *line == '\0' )
         break;

      ptr = str_dup( line, item->count );

      {
         char *ptr1;

         /*
          * Convert header keyword to lowercase.
          */

         for( ptr1 = line; *ptr1; ++ptr1 )
            if ( *ptr1 == ':' )
               break;
            else if ( *ptr1 >= 65 && *ptr1 <= 90 )
               *ptr1 += 32;
      }

      /*
       * CGI programs must not create headers larger than the ccb->buffer
       * - 1000.  This give us space to supplement the header without
       * the possibility of buffer overrun.
       */

      if (( item->size += item->count ) > IOBUFSIZE - 1000 )
      {
         free( ptr );
         remove_conn( item );
         return;
      }

      STACK_PUSH( item->working, ptr )
      if ( alloc_err )
      {
         remove_conn( item );
         return;
      }

      if ( ! item->type_found )
      {
         char *ptr1, *ptr2;

         ptr1 = line;
         ptr2 = "content-type:";

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr2 )
         {
            item->type_found = 1;
            continue;
         }
      }

      if ( ! item->location_found )
      {
         char *ptr1, *ptr2;

         ptr1 = line;
         ptr2 = "location:";

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr2 )
         {
            item->location_found = 1;
            continue;
         }
      }

      if ( ! item->date_found )
      {
         char *ptr1, *ptr2;

         ptr1 = line;
         ptr2 = "date:";

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr2 )
         {
            item->date_found = 1;
            continue;
         }
      }

      if ( ! item->length_found )
      {
         if (( matches = extract_header( "content-length:", line )) == NULL )
         {
            if ( alloc_err )
            {
               remove_conn( item );
               return;
            }
         }
         else
         {
            item->length_found = 1;
            item->length = matches[ 0 ];
         }
      }
   }

   item->written = 0;

   if ( item->location_found )
      item->next = HEADER;
   else if ( ! item->verno || item->length != NULL )
      item->next = NOCHUNK;
   else
      item->next = CHUNK;

   item->state = HEADER;

   notready = 0;

   for( r = 0; r < item->working->used; ++r )
   {
      for( ptr = item->working->values[ r ]; *ptr; ++ptr )
         ++notready;
      notready += + 2;
   }

   if (( ptr = memory( ++notready )) == NULL )
   {
      remove_conn( item );
      return;
   }

   *ptr = '\0';

   while(( line = ( char *)stack_pop( item->working )) != NULL )
   {
      strcat( ptr, line );
      strcat( ptr, "\r\n" );
      free( line );
   }

   item->count = snprintf( item->buffer, IOBUFSIZE, "%s%s%s%s%s%s%s%s%s\r\n",
             item->version,
             ( item->location_found ? " 302 Found\r\n" : " 200 OK\r\n" ),
             server_header,
             ( item->date_found ? "" : date_header() ),
             ( item->location_found ? "Content-Length: 0\r\n" : "" ),
             ( ! item->type_found ? default_content_type : "" ),
             ( ! item->location_found && item->next == CHUNK ? "Transfer-Encoding: chunked\r\n" : "" ),
             ( item->close ? "Connection: close\r\n" : "" ),
             ptr );

   free( ptr );
   non_blocking( item->sock );
   flip_events( item );
}

void accept_connection( mfd )
{
   struct sockaddr_storage addr;
   char address[ 128 ];
   const char *ptr = NULL;
   static socklen_t len = sizeof( struct sockaddr_storage );
   int conn;

   conn = accept( mfd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno != ECONNABORTED && errno != EWOULDBLOCK && errno != EAGAIN )
         syslog( LOG_ERR, "accept(): %m" );

      return;
   }

   ++active;
   non_blocking( conn );
   close_on_exec( conn );

   if ( addr.ss_family == AF_INET6 )
      ptr = inet_ntop( AF_INET6, &(( struct sockaddr_in6 *)&addr )->sin6_addr,
                       address, sizeof( address ));
   else if ( addr.ss_family == AF_INET )
      ptr = inet_ntop( AF_INET, &(( struct sockaddr_in *)&addr )->sin_addr,
                       address, sizeof( address ));

   if ( ptr == NULL )
      strcpy( address, "unknown" );

   add_conn( conn, address );
}

void remove_events( int n )
{
   while( ++n < out )
      if ( outqueue[ n ].ident == closed )
         outqueue[ n ].ident = -1;
}

/*
 * If a timer fires on a descriptor and that descriptor becomes readable
 * after the last invocation of kevent(), we could get the both events in
 * the queue, on the next invocation.  If the read event was earlier in the
 * queue, and in processing the read event, it was decided to close the
 * connection, its ccb would be freed with a call to remove_conn().  When
 * the timer event was encountered a second freeing of the ccb would occur.
 * If the events were placed in the queue in the opposite order, then
 * processing the timer event would call remove_conn(), and then the read
 * event handler would be dispatched with a pointer to the freed ccb.
 * Both situations will crash the server.

 * Our solution is to modify the output queue every time a connection is
 * dropped during the processing of a read, write, or timer event.  We make
 * sure any other events for the same descriptor further along in the queue
 * are ignored by setting the ident member of the appropriate struct kevent
 * to -1.

 * kevent() returns -1 when there is no room to create an error event in
 * the output queue.  We treat this case the same as an error event in the
 * output queue, and just move on.  Errors are generated from duplicate
 * delete timer events in the input queue, and the occasional failure to
 * create a read or a write event on a connection unexpectedly dropped by
 * the client or server.  When kevent() does return -1, the output queue
 * may contain events, but we don't know how many, so we can't service
 * them.  If they are errors, we don't care.  If they are IO events, they
 * are level-triggered, so we will get those events again.  If they are
 * timer events, they will be generated again, for timers are periodic.
 * The worst that can happen is that a timeout will take longer than
 * otherwise to occur.  No events are lost.

 * Because we the listening socket can only appear once in the output queue,
 * the rate at which we accept new connections is inversely proportional to
 * the number of active connections.
 */

void process_clients()
{
   int n, kq;
   struct ccb *item;

   if (( kq = kqueue()) < 0 )
   {
      syslog( LOG_ERR, "kqueue(): %m" );
      exit( 1 );
   }

   close_on_exec( kq );
   ev_set( fd, EVFILT_READ, EV_ADD | EV_ENABLE, NULL );

   for( ; ; )
   {
      if ( terminate )
      {
         if ( fd > 0 )
         {
            close( fd );
            fd = -1;
         }

         if ( ! active )
            exit( 0 );
      }

      out = kevent( kq, inqueue, in, outqueue, QLEN, NULL );

      in = 0;

      if ( out <= 0 )
         continue;

      for( n = 0; n < out; ++n )
      {
         if ( ( int )outqueue[ n ].ident < 0 || outqueue[ n ].flags & EV_ERROR )
            continue;

         closed = 0;
         item = ( struct ccb *)outqueue[ n ].udata;

         if ( outqueue[ n ].filter == EVFILT_TIMER )
            remove_conn( item );
         else if ( outqueue[ n ].ident == fd )
            accept_connection( fd );
         else
            item->state( item );

         if ( closed )
            remove_events( n );
      }
   }
}

int fork_children( instances )
{
   pid_t pid;

   if (( children = memory( instances * sizeof( pid_t ))) == NULL )
      exit( 1 );

   while( instances-- )
      switch(( pid = fork() ))
      {
         case -1:
            syslog( LOG_ERR, "fork(): %m" );
            return 1;

         case 0:
            free( children );
            children = NULL;
            return 0;

         default:
            children[ instances ] = pid;
      }

   return 1;
}

void open_request_log()
{
   if ( logfile == NULL )
      return;

   if (( logfd = open( logfile, O_WRONLY | O_CREAT | O_APPEND,
                       S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
      syslog( LOG_ERR, "open(): %m" );
   else
      close_on_exec( logfd );
}

void change_directory()
{
   if ( chdir( root ) < 0 )
   {
      syslog( LOG_ERR, "chdir(): %m" );
      exit( 1 );
   }
}

/*
 * Master process gets SIGCHLD should server process exit prematurely.
 */

int restart_children()
{
   int i;
   pid_t pid;

   while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )

      for( i = 0; i < instances; ++i )
         if ( children[ i ] == pid )

            switch(( pid = fork() ))
            {
               case -1:
                  syslog( LOG_ERR, "fork(): %m" );
                  break;

               case 0:
                  signal( SIGHUP, SIG_IGN );
                  signal( SIGCHLD, SIG_IGN );
                  return 1;

               default:
                  children[ i ] = pid;
            }

   return 0;
}

void sigchld_handler( int signo )
{
   sigchld = 1;
}

void wait_for_signals()
{
   sigset_t sigset;

   /*
    * SIGHUP and SIGCHLD are handled only by master process.
    */

   signal( SIGHUP, sighup_handler );
   signal( SIGCHLD, sigchld_handler );

   if ( sigprocmask( 0, NULL, &sigset ) < 0 )
      syslog( LOG_ERR, "sigprocmask(): %m" );

   for( ; ; )
   {
      sigsuspend( &sigset );

      if ( restart )
         restart_drood();
      else if ( terminate )
         exit( 0 );
      else if ( sigchld )
      {
         sigchld = 0;

         if ( restart_children() )
            return;
      }
   }
}

void init_stat_cache()
{
   int i;
   struct sc *ptr;

   for( ptr = stat_cache, i = 0; i < CACHE_SIZE + 1; ++i, ++ptr )
      bzero( ptr, sizeof( struct sc ));
}

int main( int argc, char **argv )
{
   set_options( argc, argv );
   set_signals();
   set_types();

   openlog( "drood", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();

   if ( ! testing && fork_children( instances ))
      wait_for_signals();

   open_request_log();
   change_directory();

   change_identity();
   init_stat_cache();
   process_clients();

   return 0;
}
