/*
 * Plunger Copyright (c) 2010, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Multi-process generic SCGI application server.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "1.0"
#define PIDFILE "/var/run/plunger.pid"

#define NOT_BUSY 0
#define BUSY 1
#define EMPTY -1

/*
 * We will only store 20 Enviroment variables and values for each
 * request.

 * We will not maintain more than ten running responder programs.
 */

#define MAX_ENV 40
#define MAX_RESPONDERS 10

extern char *optarg;

int fd = -1, testing = 0, backlog = 1000, min_slaves = 3, max_body = 10000,
   logging = 0, max_slaves = 25, slave_no = -1, header_len = 0;

volatile int killed = 0;

pid_t *slaves = NULL;

char *interface = "", *port = "4000", *grp = "nobody", *user = "nobody",
   *busymap = NULL, *header = NULL, *body = NULL, *root_dir = NULL,
   *env[ MAX_ENV + 1 ];

struct passwd *passwd;
struct group *group;

int stack_inc = 128;

void *memory( int size );

struct responder {
   FILE *fp;
   char *name;
} responders[ MAX_RESPONDERS + 1 ];

struct string
{
   int free, used;
   char *top;
   char *str;
};

void string_free( struct string *s )
{
   free( s->str );
   free( s );
}

struct string *make_string()
{
   struct string *s;

   if (( s = ( struct string *)memory( sizeof( struct string ))) != NULL )
   {
      if (( s->str = ( char *)memory( stack_inc + 1 )) != NULL )
      {
         *s->str = '\0';
         s->free = stack_inc;
         s->used = 0;
         s->top = s->str;
      }
   }

   return s;
}

int string_append( struct string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + stack_inc );

      if ( s->str == NULL )
      {
         syslog( LOG_ERR, "realloc(): %m" );
         return 1;
      }

      /* Leave room for end-of-string sentinel */

      s->free = stack_inc;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';

   return 0;
}

void *memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "plunger: malloc(): %s\n", strerror( errno ));
   }

   return ptr;
}

char *str_dup( char *str, int len )
{
   char *ptr;

   ptr = ( char *)memory( len + 1 );
   bcopy( str, ptr, len );
   ptr[ len ] = '\0';

   return ptr;
}

/*
 * If we are the master process, we send our children SIGKILLS so they
 * will exit immediately.  SIGTERM is used by manage_slaves() to tell
 * slaves to exit at their next convenience.

 * If we are a slave process, we set the killed flag, and when we are
 * finished servicing any current connection, we will exit.  It would be
 * nice to unlink our PIDFILE here, but we no longer are root and so
 * cannot access /var/run/
 */

void signal_handler( int signo )
{
   int i;

   if ( slave_no >= 0 )
      killed = 1;
   else
   {
      for( i = 0; i < max_slaves; ++i )
         if ( slaves[ i ] )
            kill( slaves[ i ], SIGKILL );

      exit( 0 );
   }
}

/*
 * x - Do not become daemon.  Run single process iterative server.

 * m - max slaves
 * n - min free slaves
 * p - port
 * r - root directory
 * i - interface
 * u - user
 * g - group
 * b - max size of request bodies
 */

void set_options( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xr:m:b:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'b':
            max_body = atoi( optarg );
            break;

         case 'r':
            root_dir = optarg;
            break;

         case 'g':
            grp = optarg;
            break;

         case 'i':
            interface = optarg;
            break;

         case 'm':
            max_slaves = atoi( optarg );
            break;

         case 'n':
            min_slaves = atoi( optarg );
            break;

         case 'p':
            port = optarg;
            break;

         case 'u':
            user = optarg;
            break;

         case 'x':
            testing = 1;
      }

   if ( root_dir == NULL )
   {
      fprintf( stderr, "plunger: root directory not set\n" );
      exit( 1 );
   }

   if ( chdir( root_dir ) < 0 )
   {
      fprintf( stderr, "plunger: chdir( %s ): %s\n", root_dir,
               strerror( errno ));
      exit( 1 );
   }

   if ( min_slaves <= 0 )
   {
      fprintf( stderr, "plunger: min slaves <= 0: %d\n", min_slaves );
      exit( 1 );
   }

   if ( max_slaves <= 0 )
   {
      fprintf( stderr, "plunger: max slaves <= 0: %d\n", max_slaves );
      exit( 1 );
   }

   if ( min_slaves > max_slaves )
   {
      fprintf( stderr, "plunger: min slaves > max_slaves: %d > %d\n",
               min_slaves, max_slaves );
      exit( 1 );
   }

   if ( max_body <= 0 )
   {
      fprintf( stderr, "plunger: max_body <= 0: %d\n", max_body );
      exit( 1 );
   }

   if (( passwd = getpwnam( user )) == NULL )
   {
      fprintf( stderr, "plunger: user \"%s\" does not exist\n", user );
      exit( 1 );
   }

   if (( group = getgrnam( grp )) == NULL )
   {
      fprintf( stderr, "plunger: group \"%s\" does not exist\n", grp );
      exit( 1 );
   }
}

void become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "plunger: fork(): %s\n", strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if ( ! testing )
   {
      if (( file = open( PIDFILE, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
         syslog( LOG_WARNING, "open(): %m" );
      else
      {
         char buffer[ 16 ];

         snprintf( buffer, sizeof( buffer ), "%d", getpid() );
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   /*
    * If responder children create files, this will make those files not
    * world-readable by default.
    */

   umask( 2 );
}

void change_identity()
{
   if ( setgid( group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "setgid(): %m" );
      exit( 1 );
   }

   if ( setuid( passwd->pw_uid ) < 0 )
   {
      syslog( LOG_ERR, "setuid(): %m" );
      exit( 1 );
   }
}

void close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

void start_listening()
{
   struct addrinfo hints, *res;
   int result;

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *interface ? interface : NULL ), port,
                               &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can have more than one
    * server process, if necessary.
    */

   if ( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *interface &&
        setsockopt( fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( fd, backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   close_on_exec( fd );
}

/*
 * If we are a slave, and we receive a SIGTERM, we will exit at our next
 * convenience, either the next time accept_connection() is called, after
 * we have closed the previous connection, but before we accept another, or
 * if we are already blocked in accept(), immediately, when accept() is
 * interrupted by the signal.
 */

void accept_connection()
{
   struct sockaddr_storage addr;
   socklen_t len;
   int conn;

   fclose( stdin );
   fclose( stdout );

AGAIN:
   if ( killed )
      exit( 0 );

   len = sizeof( struct sockaddr_storage );
   conn = accept( fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno == EINTR )
         goto AGAIN;

      syslog( LOG_ERR, "accept(): %m" );
      exit( 1 );
   }

   if (( stdin = fdopen( conn, "r" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }

   if (( conn = dup( conn )) < 0 )
   {
      syslog( LOG_ERR, "dup(): %m" );
      exit( 1 );
   }

   if (( stdout = fdopen( conn, "w" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }
}

char *find_next( char *ptr )
{
   while( --header_len )
      if ( ! *ptr++ )
         break;

   if ( ! header_len )
      return NULL;

   return ptr;
}

int init_env( int len )
{
   char *ptr;
   int n;

   n = 0;

   for( ptr = header; ptr != NULL; ptr = find_next( ptr ))
   {
      env[ n++ ] = ptr;

      if ( n == MAX_ENV )
         break;
   }

   if ( n % 2 )
      return 1;

   env[ n ] = NULL;

   return 0;
}

int read_header()
{
   char *ptr, buffer[ 64 ];
   int c;

   for( ptr = buffer; ; ++ptr )
   {
      if (( c = getchar()) == EOF )
         return 1;

      if ( c == ':' )
      {
         *ptr = '\0';
         break;
      }

      *ptr = c;
   }

   if (( header_len = atoi( buffer )) <= 0 )
      return 1;

   if (( header = memory( header_len )) == NULL )
      return 1;

   if ( fread( header, 1, header_len, stdin ) < c )
   {
      free( header );
      header = NULL;
      return 1;
   }

   if ( getchar() != ',' )
   {
      free( header );
      header = NULL;
      return 1;
   }

   return init_env( header_len );
}

int read_body()
{
   char *ptr;
   int count;

   for( ptr = header; *ptr; ++ptr )
      ;

   if ( !( count = atoi( ++ptr )))
      return 0;

   /*
    * We drop connections where the client is attempting
    * to give us a request body greater than max_body characters
    * in length.
    */

   if ( count < 0 || count > max_body )
   {
      free( header );
      header = NULL;
      return 1;
   }

   if (( body = memory( count + 1 )) == NULL )
   {
      free( header );
      header = NULL;
      return 1;
   }

   if ( fread( body, 1, count, stdin ) < count )
   {
      free( header );
      free( body );
      header = body = NULL;
      return 1;
   }

   body[ count ] = '\0';

   return 0;
}

char *get_env( char *key )
{
   char **ptr;

   for( ptr = env; *ptr != NULL; ptr += 2 )
      if ( ! strcmp( *ptr, key ))
         return *( ptr + 1 );

   return NULL;
}

FILE *fork_responder( char *script )
{
   FILE *fp;
   int pfd[ 2 ], pid;

   if ( pipe( &pfd[ 0 ] ) < 0 )
   {
      syslog( LOG_ERR, "pipe(): %m" );
      return NULL;
   }

   switch(( pid = fork() ))
   {
      case -1:
         close( pfd[ 0 ] );
         close( pfd[ 1 ] );
         syslog( LOG_ERR, "fork(): %m" );
         return NULL;

      case 0:
         /*
          * Closes client socket in child.  Listening socket
          * had been set close-on-exec.
          */

         fclose( stdin );
         fclose( stdout );

         if ( dup2( pfd[ 1 ], 0 ) < 0 )
         {
            syslog( LOG_ERR, "dup2(): %m" );
            _exit( 1 );
         }

         if ( dup2( pfd[ 1 ], 1 ) < 0 )
         {
            syslog( LOG_ERR, "dup2(): %m" );
            _exit( 1 );
         }

         close( pfd[ 0 ] );
         close( pfd[ 1 ] );

         closelog();
         execl( script, script, NULL );
         syslog( LOG_ERR, "excel(): %m" );
         _exit( 1);

      default:
         close( pfd[ 1 ] );
   }

   /*
    * So the next time we fork() a responder, it won't inherit open
    * descriptors connected to other responders.
    */

   close_on_exec( pfd[ 0 ] );

   if (( fp = fdopen( pfd[ 0 ], "r+" )) == NULL )
   {
      close( pfd[ 0 ] );
      syslog( LOG_ERR, "fdopen(): %m" );
   }

   return fp;
}

void abandon_responder( struct responder *responder )
{
   if ( responder->fp != NULL )
   {
      fclose( responder->fp );
      responder->fp = NULL;
   }

   if ( responder->name != NULL )
   {
      free( responder->name );
      responder->name = NULL;
   }

   while( ( ++responder )->fp != NULL )
      *( responder - 1 ) = *responder;
}

int exchange_data ( FILE *file )
{
   char **ptr, *last, b0[ 1024 ], *current, b1[ 1024 ], *result;

   for( ptr = env; *ptr; ptr += 2 )
      if ( fprintf( file,  "%s:%s\n", *ptr, *( ptr + 1 )) < 0 )
      {
         syslog( LOG_ERR, "fprintf(): %m" );
         return 1;
      }

   if ( fputs( "\n", file ) == EOF )
   {
      syslog( LOG_ERR, "fputs(): %m" );
      return 1;
   }

   if ( body != NULL && fputs( body, file ) == EOF )
   {
      syslog( LOG_ERR, "fputs(): %m" );
      return 1;
   }

   if ( fputs( "\n__END__\n", file ) == EOF )
   {
      syslog( LOG_ERR, "fputs(): %m" );
      return 1;
   }

   fflush( file );

   last = b0;
   current = b1;
   *last = '\0';

   for( ; ; )
   {
      result = fgets( current, 1024, file );

      if ( result != NULL )
      {
         if ( ! strcmp( "__END__\n", current ))
         {
            if ( *last )
            {
               /*
                * Remove the extra terminating newline from the response.
                */

               last[ strlen( last ) - 1 ] = '\0';

               if ( fputs( last, stdout ) == EOF )
               {
                  syslog( LOG_ERR, "fputs(): %m" );
                  return 1;
               }
            }

            return 0;
         }

         if ( fputs( last, stdout ) == EOF )
         {
            syslog( LOG_ERR, "fputs(): %m" );
            return 1;
         }

         result = last;
         last = current;
         current = result;

         continue;
      }

      if ( feof( file ))
      {
         syslog( LOG_ERR, "premature EOF from responder" );
         return 1;
      }

      if ( ferror( file ))
      {
         syslog( LOG_ERR, "fgets(): %m" );
         return 1;
      }
   }

   /* not reached */

   return 0;
}

void process_request( char *script )
{
   struct responder *responder;
   int i;

   for( i = 0, responder = responders; responder->name != NULL; ++responder, ++i )
      if ( ! strcmp( responder->name, script ))
         break;

   if ( responder->name == NULL )
   {
      if ( i == MAX_RESPONDERS )
      {
         syslog( LOG_ERR, "MAX_RESPONDERS reached" );
         return;
      }

      if (( responder->fp = fork_responder( script )) == NULL )
         return;

      if (( responder->name = str_dup( script, strlen( script ))) == NULL )
         abandon_responder( responder );
   }

   if ( exchange_data( responder->fp ))
      abandon_responder( responder );
}

void service_connections()
{
   char *ptr;
   struct string *path;
   int len;

   if (( path = make_string()) == NULL )
      exit( 1 );

   for( ptr = root_dir; *ptr; ++ptr )
      if ( string_append( path, *ptr ))
         exit( 1 );

   len = path->used;

   /*
    * We want write errors when responders exit, not signals.
    */

   signal( SIGPIPE, SIG_IGN );

   for( ; ; )
   {
      busymap[ slave_no ] = NOT_BUSY;
      accept_connection();
      busymap[ slave_no ] = BUSY;

      if ( read_header() )
         continue;

      if ( read_body() )
         continue;

      if (( ptr = get_env( "SCRIPT_NAME" )) == NULL || ! *ptr )
         syslog( LOG_ERR, "SCRIPT_NAME not set!" );
      else
      {
         while( *ptr )
            if ( string_append( path, *ptr++ ))
               goto NEXT;

         process_request( path->str );

   NEXT:
         path->free = ( path->used + path->free ) - len;
         path->used = len;
         path->top = &path->str[ len ];
         *path->top = '\0';
      }

      free( header );
      if ( body != NULL )
         free( body );

      header = body = NULL;
   }
}

void init_busymap()
{
   int i;

   if (( busymap = mmap( 0, max_slaves, PROT_WRITE | PROT_READ,
                            MAP_ANON | MAP_SHARED, -1, 0 )) == NULL )
   {
      syslog( LOG_ERR, "mmap(): %m" );
      exit( 1 );
   }

   if (( slaves = memory( sizeof( pid_t ) * max_slaves )) == NULL )
      exit( 1 );

   for( i = 0; i < max_slaves; ++i )
   {
      busymap[ i ] = EMPTY;
      slaves[ i ] = 0;
   }
}

void fork_slave( int i )
{
   pid_t pid;

   switch(( pid = fork()))
   {
      case -1:
         syslog( LOG_ERR, "fork(): %m" );
         exit( 1 );

      case 0:
         slave_no = i;
         service_connections();
         exit( 0 );

      default:
         slaves[ i ] = pid;
   }
}

void manage_slaves()
{
   pid_t pid;
   int i, idle, last_idle;

   for( ; ; )
   {
      sleep( 2 );
      idle = 0;
      last_idle = -1;

      for( i = 0; i < max_slaves; ++i )
         if ( busymap[ i ] == NOT_BUSY )
         {
            last_idle = i;
            ++idle;
         }

      if ( idle > min_slaves )
      {
         kill( slaves[ last_idle ], SIGTERM );
         --idle;
      }

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < max_slaves; ++i )
            if ( pid == slaves[ i ] )
            {
               slaves[ i ] = 0;
               busymap[ i ] = EMPTY;
               break;
            }
      }

      while( idle < min_slaves )
      {
         for( i = 0; i < max_slaves; ++i )
            if ( ! slaves[ i ] )
            {
               fork_slave( i );
               ++idle;
               break;
            }

         if ( i == max_slaves )
            break;
      }
   }
}

void set_signals()
{
   struct sigaction sigact;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * We ignore everything but SIGTERM when ! testing.  Otherwise, we
    * catch all the signals, so we may kill our children before exiting.
    * SIGTERM is set to interrupt system calls.
    */

   sigact.sa_handler = signal_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGTERM, &sigact, NULL ) < 0 )
      fprintf( stderr, "plunger: sigaction: %s\n", strerror( errno ));

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( testing ? signal_handler : SIG_IGN ));
}

void fork_slaves()
{
   int i;

   for( i = 0; i < min_slaves; ++i )
      fork_slave( i );
}

int main( int argc, char **argv )
{
   int i;

   set_options( argc, argv );
   set_signals();

   openlog( "plunger", LOG_PID, LOG_DAEMON );
   logging = 1;

   if ( !testing )
      become_daemon();

   start_listening();
   change_identity();
   init_busymap();

   for( i = 0; i <= MAX_RESPONDERS; ++i )
   {
      responders[ i ].fp = NULL;
      responders[ i ].name = NULL;
   }

   if ( testing )
      service_connections();
   else
   {
      fork_slaves();
      manage_slaves();
   }

   return 0;
}
